from __future__ import annotations

import collections
import traceback
from dataclasses import dataclass, field
from functools import partial
from math import floor
from time import perf_counter
from typing import Callable, DefaultDict, Dict, List, Optional, Sequence, Tuple, TypeVar

from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    MetricModelReference,
    SemanticModelElementReference,
    SemanticModelReference,
)
from dbt_semantic_interfaces.validations.validator_helpers import (
    FileContext,
    MetricContext,
    SavedQueryContext,
    SavedQueryElementType,
    SemanticManifestValidationResults,
    SemanticModelContext,
    SemanticModelElementContext,
    SemanticModelElementType,
    ValidationContext,
    ValidationError,
    ValidationIssue,
    ValidationWarning,
)
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.specs.dunder_column_association_resolver import DunderColumnAssociationResolver
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.measure_spec import MeasureSpec
from metricflow_semantics.specs.spec_set import InstanceSpecSet
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.source_node import SourceNodeBuilder
from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
from metricflow.dataset.convert_semantic_model import SemanticModelToDataSetConverter
from metricflow.dataset.dataset_classes import DataSet
from metricflow.engine.metricflow_engine import MetricFlowEngine, MetricFlowExplainResult, MetricFlowQueryRequest
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlPlanConverter
from metricflow.protocols.sql_client import SqlClient


@dataclass
class QueryRenderingTools:
    """General tools that data warehosue validations use for rendering the validation queries.

    This is necessary for the validation steps that generate raw partial queries against the individual semantic models
    (e.g., selecting a single dimension column).
    """

    semantic_manifest_lookup: SemanticManifestLookup
    source_node_builder: SourceNodeBuilder
    converter: SemanticModelToDataSetConverter
    plan_converter: DataflowToSqlPlanConverter

    def __init__(self, manifest: SemanticManifest) -> None:  # noqa: D107
        self.semantic_manifest_lookup = SemanticManifestLookup(semantic_manifest=manifest)
        self.source_node_builder = SourceNodeBuilder(
            column_association_resolver=DunderColumnAssociationResolver(),
            semantic_manifest_lookup=self.semantic_manifest_lookup,
        )
        self.converter = SemanticModelToDataSetConverter(column_association_resolver=DunderColumnAssociationResolver())
        self.plan_converter = DataflowToSqlPlanConverter(
            column_association_resolver=DunderColumnAssociationResolver(),
            semantic_manifest_lookup=self.semantic_manifest_lookup,
        )
        self.node_resolver = DataflowPlanNodeOutputDataSetResolver(
            column_association_resolver=DunderColumnAssociationResolver(),
            semantic_manifest_lookup=self.semantic_manifest_lookup,
        )


@dataclass
class DataWarehouseValidationTask:
    """Dataclass for defining a task to be used in the DataWarehouseModelValidator."""

    query_and_params_callable: Callable[[], Tuple[str, SqlBindParameterSet]]
    error_message: str
    description: str
    context: Optional[ValidationContext] = None
    on_fail_subtasks: List[DataWarehouseValidationTask] = field(default_factory=lambda: [])


LinkableInstanceSpecT = TypeVar("LinkableInstanceSpecT", bound=LinkableInstanceSpec)


class DataWarehouseTaskBuilder:
    """Task builder for standard data warehouse validation tasks."""

    @staticmethod
    def _remove_entity_link_specs(specs: Tuple[LinkableInstanceSpecT, ...]) -> Tuple[LinkableInstanceSpecT, ...]:
        """For the purposes of data warehouse validation, specs with entity_links are unnecesary."""
        return tuple(spec for spec in specs if not spec.entity_links)

    @staticmethod
    def _semantic_model_nodes(
        render_tools: QueryRenderingTools, semantic_model: SemanticModel
    ) -> Sequence[DataflowPlanNode]:
        """Builds and returns the SemanticModelDataSet node for the given semantic model."""
        fetched_semantic_model = render_tools.semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
            SemanticModelReference(semantic_model_name=semantic_model.name)
        )
        assert fetched_semantic_model is not None

        source_nodes = render_tools.source_node_builder.create_from_data_sets(
            (render_tools.converter.create_sql_source_data_set(fetched_semantic_model),)
        ).source_nodes_for_metric_queries

        assert len(source_nodes) >= 1
        return source_nodes

    @staticmethod
    def renderize(
        sql_client: SqlClient, plan_converter: DataflowToSqlPlanConverter, plan_id: str, nodes: FilterElementsNode
    ) -> Tuple[str, SqlBindParameterSet]:
        """Generates a sql query plan and returns the rendered sql and bind_parameter_set."""
        conversion_result = plan_converter.convert_to_sql_plan(
            sql_engine_type=sql_client.sql_engine_type,
            dataflow_plan_node=nodes,
        )
        sql_plan = conversion_result.sql_plan

        rendered_plan = sql_client.sql_plan_renderer.render_sql_plan(sql_plan)
        return (rendered_plan.sql, rendered_plan.bind_parameter_set)

    @classmethod
    def gen_semantic_model_tasks(
        cls, manifest: SemanticManifest, semantic_model_filters: Optional[Sequence[str]] = None
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the semantic models of the manifest."""
        tasks: List[DataWarehouseValidationTask] = []
        for semantic_model in manifest.semantic_models:
            if semantic_model_filters is not None and semantic_model.name not in semantic_model_filters:
                continue
            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        lambda name=semantic_model.node_relation.relation_name: (
                            f"SELECT * FROM {name}",
                            SqlBindParameterSet(),
                        )
                    ),
                    context=SemanticModelContext(
                        file_context=FileContext.from_metadata(metadata=semantic_model.metadata),
                        semantic_model=SemanticModelReference(semantic_model_name=semantic_model.name),
                    ),
                    error_message=f"Unable to access semantic model `{semantic_model.name}` in data warehouse",
                    description=f"Validating semantic_model {semantic_model.name}",
                )
            )

        return tasks

    @classmethod
    def gen_dimension_tasks(
        cls,
        manifest: SemanticManifest,
        sql_client: SqlClient,
        semantic_model_filters: Optional[Sequence[str]] = None,
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the dimensions of the manifest.

        The high level tasks returned are "short cut" queries which try to
        query all the dimensions for a given semantic model. If that query fails,
        one or more of the dimensions is incorrectly specified. Thus if the
        query fails, there are subtasks which query the individual dimensions
        on the semantic model to identify which have issues.
        """
        render_tools = QueryRenderingTools(manifest=manifest)

        tasks: List[DataWarehouseValidationTask] = []
        for semantic_model in manifest.semantic_models:
            if not semantic_model.dimensions:
                continue

            if semantic_model_filters is not None and semantic_model.name not in semantic_model_filters:
                continue

            source_node = cls._semantic_model_nodes(render_tools=render_tools, semantic_model=semantic_model)[0]

            semantic_model_sub_tasks: List[DataWarehouseValidationTask] = []
            dataset = render_tools.converter.create_sql_source_data_set(semantic_model)

            dimension_specs = DataWarehouseTaskBuilder._remove_entity_link_specs(
                dataset.instance_set.spec_set.dimension_specs
            )

            spec_filter_tuples = []
            for spec in dimension_specs:
                spec_filter_tuples.append(
                    (
                        spec,
                        FilterElementsNode.create(
                            parent_node=source_node, include_specs=InstanceSpecSet(dimension_specs=(spec,))
                        ),
                    )
                )

            time_dimension_specs = DataWarehouseTaskBuilder._remove_entity_link_specs(
                dataset.instance_set.spec_set.time_dimension_specs
            )
            for spec in time_dimension_specs:
                spec_filter_tuples.append(
                    (
                        spec,
                        FilterElementsNode.create(
                            parent_node=source_node, include_specs=InstanceSpecSet(time_dimension_specs=(spec,))
                        ),
                    )
                )

            for spec, filter_elements_node in spec_filter_tuples:
                semantic_model_sub_tasks.append(
                    DataWarehouseValidationTask(
                        query_and_params_callable=partial(
                            cls.renderize,
                            sql_client=sql_client,
                            plan_converter=render_tools.plan_converter,
                            plan_id=f"{semantic_model.name}_dim_{spec.element_name}_validation",
                            nodes=filter_elements_node,
                        ),
                        context=SemanticModelElementContext(
                            file_context=FileContext.from_metadata(metadata=semantic_model.metadata),
                            semantic_model_element=SemanticModelElementReference(
                                semantic_model_name=semantic_model.name, element_name=spec.element_name
                            ),
                            element_type=SemanticModelElementType.DIMENSION,
                        ),
                        error_message=f"Unable to query dimension `{spec.qualified_name}` on semantic model `{semantic_model.name}` in data warehouse",
                        description=f"Validating dimension `{spec.qualified_name}` in semantic_model `{semantic_model.name}`",
                    )
                )

            filter_elements_node = FilterElementsNode.create(
                parent_node=source_node,
                include_specs=InstanceSpecSet(
                    dimension_specs=dimension_specs,
                    time_dimension_specs=time_dimension_specs,
                ),
            )
            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        cls.renderize,
                        sql_client=sql_client,
                        plan_converter=render_tools.plan_converter,
                        plan_id=f"{semantic_model.name}_all_dimensions_validation",
                        nodes=filter_elements_node,
                    ),
                    context=SemanticModelContext(
                        file_context=FileContext.from_metadata(metadata=semantic_model.metadata),
                        semantic_model=SemanticModelReference(semantic_model_name=semantic_model.name),
                    ),
                    error_message=f"Failed to query dimensions in data warehouse for semantic model `{semantic_model.name}`",
                    on_fail_subtasks=semantic_model_sub_tasks,
                    description=f"Validating all dimensions in semantic_model `{semantic_model.name}`",
                )
            )
        return tasks

    @classmethod
    def gen_entity_tasks(
        cls,
        manifest: SemanticManifest,
        sql_client: SqlClient,
        semantic_model_filters: Optional[Sequence[str]] = None,
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the entities of the manifest.

        The high level tasks returned are "short cut" queries which try to
        query all the entities for a given semantic model. If that query fails,
        one or more of the entities is incorrectly specified. Thus if the
        query fails, there are subtasks which query the individual entities
        on the semantic model to identify which have issues.
        """
        render_tools = QueryRenderingTools(manifest=manifest)

        tasks: List[DataWarehouseValidationTask] = []
        for semantic_model in manifest.semantic_models:
            if not semantic_model.entities:
                continue
            if semantic_model_filters is not None and semantic_model.name not in semantic_model_filters:
                continue

            source_node = cls._semantic_model_nodes(render_tools=render_tools, semantic_model=semantic_model)[0]

            semantic_model_sub_tasks: List[DataWarehouseValidationTask] = []
            dataset = render_tools.converter.create_sql_source_data_set(semantic_model)
            semantic_model_specs = DataWarehouseTaskBuilder._remove_entity_link_specs(
                dataset.instance_set.spec_set.entity_specs
            )
            for spec in semantic_model_specs:
                filter_elements_node = FilterElementsNode.create(
                    parent_node=source_node, include_specs=InstanceSpecSet(entity_specs=(spec,))
                )
                semantic_model_sub_tasks.append(
                    DataWarehouseValidationTask(
                        query_and_params_callable=partial(
                            cls.renderize,
                            sql_client=sql_client,
                            plan_converter=render_tools.plan_converter,
                            plan_id=f"{semantic_model.name}_entity_{spec.element_name}_validation",
                            nodes=filter_elements_node,
                        ),
                        context=SemanticModelElementContext(
                            file_context=FileContext.from_metadata(metadata=semantic_model.metadata),
                            semantic_model_element=SemanticModelElementReference(
                                semantic_model_name=semantic_model.name, element_name=spec.element_name
                            ),
                            element_type=SemanticModelElementType.ENTITY,
                        ),
                        error_message=f"Unable to query entity `{spec.element_name}` on semantic model `{semantic_model.name}` in data warehouse",
                        description=f"Validating entity `{spec.element_name}` in semantic_model `{semantic_model.name}`",
                    )
                )

            filter_elements_node = FilterElementsNode.create(
                parent_node=source_node,
                include_specs=InstanceSpecSet(
                    entity_specs=tuple(semantic_model_specs),
                ),
            )
            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        cls.renderize,
                        sql_client=sql_client,
                        plan_converter=render_tools.plan_converter,
                        plan_id=f"{semantic_model.name}_all_entities_validation",
                        nodes=filter_elements_node,
                    ),
                    context=SemanticModelContext(
                        file_context=FileContext.from_metadata(metadata=semantic_model.metadata),
                        semantic_model=SemanticModelReference(semantic_model_name=semantic_model.name),
                    ),
                    error_message=f"Failed to query entities in data warehouse for semantic model `{semantic_model.name}`",
                    on_fail_subtasks=semantic_model_sub_tasks,
                    description=f"Validating all entities in semantic_model `{semantic_model.name}`",
                )
            )
        return tasks

    @classmethod
    def gen_measure_tasks(
        cls,
        manifest: SemanticManifest,
        sql_client: SqlClient,
        semantic_model_filters: Optional[Sequence[str]] = None,
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the measures of the manifest.

        The high level tasks returned are "short cut" queries which try to
        query all the measures for a given semantic model. If that query fails,
        one or more of the measures is incorrectly specified. Thus if the
        query fails, there are subtasks which query the individual measures
        on the semantic model to identify which have issues.
        """
        render_tools = QueryRenderingTools(manifest=manifest)

        tasks: List[DataWarehouseValidationTask] = []
        for semantic_model in manifest.semantic_models:
            if not semantic_model.measures:
                continue

            if semantic_model_filters is not None and semantic_model.name not in semantic_model_filters:
                continue

            source_nodes = cls._semantic_model_nodes(render_tools=render_tools, semantic_model=semantic_model)
            dataset = render_tools.converter.create_sql_source_data_set(semantic_model)
            semantic_model_specs = dataset.instance_set.spec_set.measure_specs

            source_node_by_measure_spec: Dict[MeasureSpec, DataflowPlanNode] = {}
            measure_specs_source_node_pair = []
            for source_node in source_nodes:
                measure_specs = render_tools.node_resolver.get_output_data_set(
                    source_node
                ).instance_set.spec_set.measure_specs
                source_node_by_measure_spec.update({measure_spec: source_node for measure_spec in measure_specs})
                measure_specs_source_node_pair.append((measure_specs, source_node))

            source_node_to_sub_task: DefaultDict[
                DataflowPlanNode, List[DataWarehouseValidationTask]
            ] = collections.defaultdict(list)
            for spec in semantic_model_specs:
                obtained_source_node = source_node_by_measure_spec.get(spec)
                assert obtained_source_node, f"Unable to find generated source node for measure: {spec.element_name}"

                filter_elements_node = FilterElementsNode.create(
                    parent_node=obtained_source_node,
                    include_specs=InstanceSpecSet(
                        measure_specs=(spec,),
                    ),
                )
                source_node_to_sub_task[obtained_source_node].append(
                    DataWarehouseValidationTask(
                        query_and_params_callable=partial(
                            cls.renderize,
                            sql_client=sql_client,
                            plan_converter=render_tools.plan_converter,
                            plan_id=f"{semantic_model.name}_measure_{spec.element_name}_validation",
                            nodes=filter_elements_node,
                        ),
                        context=SemanticModelElementContext(
                            file_context=FileContext.from_metadata(metadata=semantic_model.metadata),
                            semantic_model_element=SemanticModelElementReference(
                                semantic_model_name=semantic_model.name, element_name=spec.element_name
                            ),
                            element_type=SemanticModelElementType.MEASURE,
                        ),
                        error_message=f"Unable to query measure `{spec.element_name}` on semantic model `{semantic_model.name}` in data warehouse",
                        description=f"Validating measure `{spec.element_name}` in semantic_model `{semantic_model.name}`",
                    )
                )

            for measure_specs, source_node in measure_specs_source_node_pair:
                filter_elements_node = FilterElementsNode.create(
                    parent_node=source_node, include_specs=InstanceSpecSet(measure_specs=measure_specs)
                )
                tasks.append(
                    DataWarehouseValidationTask(
                        query_and_params_callable=partial(
                            cls.renderize,
                            sql_client=sql_client,
                            plan_converter=render_tools.plan_converter,
                            plan_id=f"{semantic_model.name}_all_measures_validation",
                            nodes=filter_elements_node,
                        ),
                        context=SemanticModelContext(
                            file_context=FileContext.from_metadata(metadata=semantic_model.metadata),
                            semantic_model=SemanticModelReference(semantic_model_name=semantic_model.name),
                        ),
                        error_message=f"Failed to query measures in data warehouse for semantic model `{semantic_model.name}`",
                        on_fail_subtasks=source_node_to_sub_task[source_node],
                        description=f"Validating all measures in semantic_model `{semantic_model.name}`",
                    )
                )
        return tasks

    @staticmethod
    def _gen_explain_query_task_query_and_params(
        mf_engine: MetricFlowEngine, mf_request: MetricFlowQueryRequest
    ) -> Tuple[str, SqlBindParameterSet]:
        explain_result: MetricFlowExplainResult = mf_engine.explain(mf_request=mf_request)
        return (
            explain_result.sql_statement.without_descriptions.sql,
            explain_result.sql_statement.without_descriptions.bind_parameter_set,
        )

    @classmethod
    def gen_metric_tasks(
        cls,
        manifest: SemanticManifest,
        sql_client: SqlClient,
        metric_filters: Optional[Sequence[str]] = None,
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the metrics of the manifest."""
        mf_engine = MetricFlowEngine(
            semantic_manifest_lookup=SemanticManifestLookup(semantic_manifest=manifest),
            sql_client=sql_client,
        )
        tasks: List[DataWarehouseValidationTask] = []
        for metric in manifest.metrics:
            if metric_filters is not None and metric.name not in metric_filters:
                continue
            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        cls._gen_explain_query_task_query_and_params,
                        mf_engine=mf_engine,
                        mf_request=MetricFlowQueryRequest.create_with_random_request_id(
                            metric_names=[metric.name], group_by_names=[DataSet.metric_time_dimension_name()]
                        ),
                    ),
                    context=MetricContext(
                        file_context=FileContext.from_metadata(metadata=metric.metadata),
                        metric=MetricModelReference(metric_name=metric.name),
                    ),
                    error_message=f"Unable to query metric `{metric.name}`",
                    description=f"Validating metric `{metric.name}`",
                )
            )
        return tasks

    @classmethod
    def gen_saved_query_tasks(
        cls,
        manifest: SemanticManifest,
        sql_client: SqlClient,
        saved_query_filters: Optional[Sequence[str]] = None,
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the saved queries of the manifest."""
        mf_engine = MetricFlowEngine(
            semantic_manifest_lookup=SemanticManifestLookup(semantic_manifest=manifest),
            sql_client=sql_client,
        )
        tasks: List[DataWarehouseValidationTask] = []
        for saved_query in manifest.saved_queries:
            if saved_query_filters is not None and saved_query.name not in saved_query_filters:
                continue
            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        cls._gen_explain_query_task_query_and_params,
                        mf_engine=mf_engine,
                        mf_request=MetricFlowQueryRequest.create_with_random_request_id(
                            saved_query_name=saved_query.name
                        ),
                    ),
                    context=SavedQueryContext(
                        file_context=FileContext.from_metadata(metadata=saved_query.metadata),
                        element_type=SavedQueryElementType.METRIC,
                        element_value=saved_query.name,
                    ),
                    error_message=f"Unable to query saved query `{saved_query.name}`",
                    description=f"Validating saved_query `{saved_query.name}`",
                )
            )
        return tasks


class DataWarehouseModelValidator:
    """A Validator for checking specific tasks for the manifest against the Data Warehouse.

    Data Warehouse Validations are validations that are done against the data
    warehouse based on the manifest configured by the user. Their purpose is to
    ensure that queries generated by MetricFlow won't fail when you go to use
    them (assuming the manifest has passed these validations before use).
    """

    def __init__(self, sql_client: SqlClient) -> None:  # noqa: D107
        self._sql_client = sql_client

    def run_tasks(
        self, tasks: List[DataWarehouseValidationTask], timeout: Optional[int] = None
    ) -> SemanticManifestValidationResults:
        """Runs the list of tasks as queries agains the data warehouse, returning any found issues.

        Args:
            tasks: A list of tasks to run against the data warehouse
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues discovered when running the passed in tasks against the data warehosue
        """
        # Used for keeping track if we go past the max time
        start_time = perf_counter()

        issues: List[ValidationIssue] = []
        # TODO: Asyncio implementation
        for index, task in enumerate(tasks):
            if timeout is not None and perf_counter() - start_time > timeout:
                issues.append(
                    ValidationWarning(
                        context=None,
                        message=f"Hit timeout before completing all tasks. Completed {index}/{len(tasks)} tasks.",
                    )
                )
                break
            try:
                (query_string, query_params) = task.query_and_params_callable()
                self._sql_client.dry_run(stmt=query_string, sql_bind_parameter_set=query_params)
            except Exception as e:
                issues.append(
                    ValidationError(
                        context=task.context,
                        message=task.error_message + f"\nReceived following error from data warehouse:\n{e}",
                        extra_detail="".join(traceback.format_tb(e.__traceback__)),
                    )
                )
                if task.on_fail_subtasks:
                    sub_task_timeout = floor(timeout - (perf_counter() - start_time)) if timeout else None
                    issues += self.run_tasks(tasks=task.on_fail_subtasks, timeout=sub_task_timeout).all_issues

        return SemanticManifestValidationResults.from_issues_sequence(issues)

    def validate_semantic_models(
        self, manifest: SemanticManifest, timeout: Optional[int] = None
    ) -> SemanticManifestValidationResults:
        """Generates a list of tasks for validating the semantic models of the model and then runs them.

        Args:
            manifest: SemanticManifest which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues discovered when running the passed in tasks against the data warehosue
        """
        tasks = DataWarehouseTaskBuilder.gen_semantic_model_tasks(manifest=manifest)
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_dimensions(
        self, manifest: SemanticManifest, timeout: Optional[int] = None
    ) -> SemanticManifestValidationResults:
        """Generates a list of tasks for validating the dimensions of the manifest and then runs them.

        Args:
            manifest: SemanticManifest which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """
        tasks = DataWarehouseTaskBuilder.gen_dimension_tasks(manifest=manifest, sql_client=self._sql_client)
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_entities(
        self, manifest: SemanticManifest, timeout: Optional[int] = None
    ) -> SemanticManifestValidationResults:
        """Generates a list of tasks for validating the entities of the manifest and then runs them.

        Args:
            manifest: SemanticManifest which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """
        tasks = DataWarehouseTaskBuilder.gen_entity_tasks(manifest=manifest, sql_client=self._sql_client)
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_measures(
        self, manifest: SemanticManifest, timeout: Optional[int] = None
    ) -> SemanticManifestValidationResults:
        """Generates a list of tasks for validating the measures of the manifest and then runs them.

        Args:
            manifest: SemanticManifest which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """
        tasks = DataWarehouseTaskBuilder.gen_measure_tasks(manifest=manifest, sql_client=self._sql_client)
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_metrics(
        self, manifest: SemanticManifest, timeout: Optional[int] = None
    ) -> SemanticManifestValidationResults:
        """Generates a list of tasks for validating the metrics of the manifest and then runs them.

        Args:
            manifest: SemanticManifest which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """
        tasks = DataWarehouseTaskBuilder.gen_metric_tasks(manifest=manifest, sql_client=self._sql_client)
        return self.run_tasks(tasks=tasks, timeout=timeout)



"""Nodes for building a dataflow plan."""

from __future__ import annotations

import functools
import logging
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import FrozenSet, Optional, Sequence, Set, Type, TypeVar

import more_itertools
from metricflow_semantics.dag.id_prefix import StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DagId, DagNode, MetricFlowDag
from metricflow_semantics.visitor import Visitable, VisitorOutputT

if typing.TYPE_CHECKING:
    from dbt_semantic_interfaces.references import SemanticModelReference
    from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec

    from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor

logger = logging.getLogger(__name__)

NodeSelfT = TypeVar("NodeSelfT", bound="DataflowPlanNode")

# Make it so that we only have to suppress errors here instead of both at the method and the class.
ComparisonAnyType = typing.Any  # type: ignore[misc]


@functools.total_ordering
@dataclass(frozen=True, eq=False)
class DataflowPlanNode(DagNode["DataflowPlanNode"], Visitable, ABC):
    """A node in the graph representation of the dataflow.

    Each node in the graph performs an operation from the data that comes from the parent nodes, and the result is
    passed to the child nodes. The flow of data starts from source nodes, and ends at sink nodes.
    """

    @property
    def _input_semantic_model(self) -> Optional[SemanticModelReference]:
        """Return the semantic model serving as direct input for this node, if one exists."""
        return None

    def as_plan(self) -> DataflowPlan:
        """Converter method for taking an arbitrary mode and producing an associated DataflowPlan.

        This is useful for doing lookups for plan-level properties at points in the call stack where we only have
        a subgraph of a complete plan. For example, the total number of nodes represented by this node and all of
        its parents would be a property of a given subgraph of the DAG. Rather than doing recursive property walks
        inside of each node, we make those properties of the DataflowPlan, and this node-level converter makes
        such properties easily accessible.
        """
        return DataflowPlan(
            sink_nodes=(self,), plan_id=DagId.from_id_prefix(id_prefix=StaticIdPrefix.DATAFLOW_PLAN_SUBGRAPH_PREFIX)
        )

    @abstractmethod
    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        """Called when a visitor needs to visit this node."""
        raise NotImplementedError

    @abstractmethod
    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:
        """Returns true if this node the same functionality as the other node.

        In other words, this returns true if  all parameters (aside from parent_nodes) are the same.
        """
        raise NotImplementedError

    @abstractmethod
    def with_new_parents(self: NodeSelfT, new_parent_nodes: Sequence[DataflowPlanNode]) -> NodeSelfT:
        """Creates a node with the same behavior as this node, but with a different set of parents.

        typing.Self would be useful here, but not available in Python 3.8.
        """
        raise NotImplementedError

    @property
    def node_type(self) -> Type:  # noqa: D102
        # TODO: Remove.
        return self.__class__

    @property
    def aggregated_to_elements(self) -> Set[LinkableInstanceSpec]:
        """Indicates that the node has been aggregated to these specs, guaranteeing uniqueness in all combinations."""
        return set()

    def __lt__(self, other: ComparisonAnyType) -> bool:  # noqa: D105
        if not isinstance(other, DataflowPlanNode):
            raise NotImplementedError

        return self.node_id < other.node_id


class DataflowPlan(MetricFlowDag[DataflowPlanNode]):
    """Describes the flow of metric data as it goes from source nodes to sink nodes in the graph."""

    def __init__(self, sink_nodes: Sequence[DataflowPlanNode], plan_id: Optional[DagId] = None) -> None:  # noqa: D107
        assert len(sink_nodes) == 1, f"Exactly 1 sink node is supported. Got: {sink_nodes}"
        super().__init__(
            dag_id=plan_id or DagId.from_id_prefix(StaticIdPrefix.DATAFLOW_PLAN_PREFIX),
            sink_nodes=tuple(sink_nodes),
        )

    @property
    def sink_node(self) -> DataflowPlanNode:  # noqa: D102
        return self._sink_nodes[0]

    @property
    def node_count(self) -> int:
        """Returns the number of nodes in the DataflowPlan."""
        return len(DataflowPlan.__all_nodes_in_subgraph(self.sink_node))

    @staticmethod
    def __all_nodes_in_subgraph(node: DataflowPlanNode) -> Sequence[DataflowPlanNode]:
        """Node accessor for retrieving a flattened sequence of all nodes in the subgraph upstream of the input node.

        Useful for gathering nodes for subtype-agnostic operations, such as common property access or simple counts.
        """
        flattened_parent_subgraphs = tuple(
            more_itertools.collapse(
                DataflowPlan.__all_nodes_in_subgraph(parent_node) for parent_node in node.parent_nodes
            )
        )
        return (node,) + flattened_parent_subgraphs

    @property
    def source_semantic_models(self) -> FrozenSet[SemanticModelReference]:
        """Return the complete set of source semantic models for this DataflowPlan."""
        return frozenset(
            [
                node._input_semantic_model
                for node in DataflowPlan.__all_nodes_in_subgraph(self.sink_node)
                if node._input_semantic_model is not None
            ]
        )



from __future__ import annotations

from dataclasses import dataclass

from metricflow_semantics.instances import InstanceSet

from metricflow.sql.sql_plan import SqlPlan


@dataclass(frozen=True)
class ConvertToSqlPlanResult:
    """Result object for returning the results of converting to a `SqlQueryPlan`."""

    instance_set: InstanceSet
    sql_plan: SqlPlan



from __future__ import annotations

from typing import Collection

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from metricflow_semantics.errors.error_classes import UnsupportedEngineFeatureError
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet
from metricflow_semantics.sql.sql_exprs import (
    SqlExtractExpression,
    SqlGenerateUuidExpression,
    SqlPercentileExpression,
    SqlPercentileFunctionType,
)
from typing_extensions import override

from metricflow.protocols.sql_client import SqlEngine
from metricflow.sql.render.expr_renderer import (
    DefaultSqlExpressionRenderer,
    SqlExpressionRenderer,
    SqlExpressionRenderResult,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlPlanRenderer


class RedshiftSqlExpressionRenderer(DefaultSqlExpressionRenderer):
    """Expression renderer for the Redshift engine."""

    sql_engine = SqlEngine.REDSHIFT

    @property
    @override
    def double_data_type(self) -> str:
        """Custom double data type for the Redshift engine."""
        return "DOUBLE PRECISION"

    @property
    @override
    def supported_percentile_function_types(self) -> Collection[SqlPercentileFunctionType]:
        return {SqlPercentileFunctionType.CONTINUOUS, SqlPercentileFunctionType.APPROXIMATE_DISCRETE}

    @override
    def visit_percentile_expr(self, node: SqlPercentileExpression) -> SqlExpressionRenderResult:
        """Render a percentile expression for Redshift."""
        arg_rendered = self.render_sql_expr(node.order_by_arg)
        params = arg_rendered.bind_parameter_set
        percentile = node.percentile_args.percentile

        if node.percentile_args.function_type is SqlPercentileFunctionType.CONTINUOUS:
            function_str = "PERCENTILE_CONT"
        elif node.percentile_args.function_type is SqlPercentileFunctionType.DISCRETE:
            raise UnsupportedEngineFeatureError(
                "Discrete percentile aggregate not supported for Redshift. Use "
                + "continuous or approximate discrete percentile in all percentile measures."
            )
        elif node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_CONTINUOUS:
            raise UnsupportedEngineFeatureError(
                "Approximate continuous percentile aggregate not supported for Redshift. Use "
                + "continuous or approximate discrete percentile in all percentile measures."
            )
        elif node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_DISCRETE:
            function_str = "APPROXIMATE PERCENTILE_DISC"
        else:
            assert_values_exhausted(node.percentile_args.function_type)

        return SqlExpressionRenderResult(
            sql=f"{function_str}({percentile}) WITHIN GROUP (ORDER BY ({arg_rendered.sql}))",
            bind_parameter_set=params,
        )

    @override
    def render_date_part(self, date_part: DatePart) -> str:
        return date_part.value

    @override
    def visit_extract_expr(self, node: SqlExtractExpression) -> SqlExpressionRenderResult:
        """Renders extract expressions with required output conversions for Redshift.

        Redshift does not have native support for the ISO standard day of week output of 1 (Monday) - 7 (Sunday).
        Instead, Redshift returns 0 (Sunday) - 6 (Monday). Therefore, we need custom rendering logic to normalize
        the return values to the ISO standard.
        """
        extract_rendering_result = super().visit_extract_expr(node)

        if node.date_part is not DatePart.DOW:
            return extract_rendering_result

        extract_stmt = extract_rendering_result.sql
        case_expr = f"CASE WHEN {extract_stmt} = 0 THEN {extract_stmt} + 7 ELSE {extract_stmt} END"

        return SqlExpressionRenderResult(
            sql=case_expr,
            bind_parameter_set=extract_rendering_result.bind_parameter_set,
        )

    @override
    def visit_generate_uuid_expr(self, node: SqlGenerateUuidExpression) -> SqlExpressionRenderResult:
        """Generates a "good enough" random key to simulate a UUID.

        NOTE: This is a temporary hacky solution as redshift does not have any UUID generation function.

        Proposed solutions that requires more thinking:
            - create a python UDF (Could we insert this without needing additional permissions?)
        """
        return SqlExpressionRenderResult(
            sql="CONCAT(CAST(RANDOM()*100000000 AS INT)::VARCHAR,CAST(RANDOM()*100000000 AS INT)::VARCHAR)",
            bind_parameter_set=SqlBindParameterSet(),
        )


class RedshiftSqlPlanRenderer(DefaultSqlPlanRenderer):
    """Plan renderer for the Redshift engine."""

    EXPR_RENDERER = RedshiftSqlExpressionRenderer()

    @property
    @override
    def expr_renderer(self) -> SqlExpressionRenderer:
        return self.EXPR_RENDERER



"""Contains input classes for the query resolver.

The naming of these classes is a little odd as they have the "For.." suffix. But using the "*ResolverInput" leads to
some confusing names like "ResolverInputForQuery" -> "QueryResolverInput". Improved naming for these classes is TBD.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Tuple, Union

from dbt_semantic_interfaces.protocols import WhereFilterIntersection
from typing_extensions import override

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.naming.metric_scheme import MetricNamingScheme
from metricflow_semantics.naming.naming_scheme import QueryItemNamingScheme
from metricflow_semantics.protocols.query_parameter import GroupByParameter, MetricQueryParameter, OrderByQueryParameter
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import (
    InputPatternDescription,
    MetricFlowQueryResolverInput,
)
from metricflow_semantics.specs.patterns.metric_pattern import MetricSpecPattern
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern


@dataclass(frozen=True)
class InvalidStringInput(MetricFlowQueryResolverInput):
    """A string input that doesn't match any of the known naming schemes."""

    input_obj: str

    @property
    @override
    def ui_description(self) -> str:
        return self.input_obj


@dataclass(frozen=True)
class ResolverInputForMetric(MetricFlowQueryResolverInput):
    """An input that describes the metrics in the query."""

    input_obj: Union[MetricQueryParameter, str]
    naming_scheme: MetricNamingScheme
    spec_pattern: MetricSpecPattern
    alias: Optional[str] = None

    @property
    @override
    def ui_description(self) -> str:
        return str(self.input_obj)

    @property
    @override
    def input_pattern_description(self) -> InputPatternDescription:
        return InputPatternDescription(
            naming_scheme=self.naming_scheme,
            spec_pattern=self.spec_pattern,
        )


@dataclass(frozen=True)
class ResolverInputForGroupByItem(MetricFlowQueryResolverInput):
    """An input that describes a group-by item in the query."""

    input_obj: Union[GroupByParameter, str]
    input_obj_naming_scheme: QueryItemNamingScheme
    spec_pattern: SpecPattern

    @property
    @override
    def ui_description(self) -> str:
        return str(self.input_obj)

    @property
    @override
    def input_pattern_description(self) -> InputPatternDescription:
        return InputPatternDescription(
            naming_scheme=self.input_obj_naming_scheme,
            spec_pattern=self.spec_pattern,
        )


@dataclass(frozen=True)
class ResolverInputForOrderByItem(MetricFlowQueryResolverInput):
    """An input that describes the ordered item.

    The challenge with order-by items is that it may not be obvious how to match an order-by item to a metric or a
    group-by item in the query. When the query inputs were always strings, this was easy because the order-by item
    could be resolved with an equality check. However, when the query inputs could be a string or a *QueryParameter
    object, the equality check is not possible. e.g. consider the case:

        group-by item: TimeDimension("creation_time"), order-by item: "creation_time".

    Instead, the approach is to resolve the metrics / group-by items into concrete spec objects, and then use the
    SpecPattern generated from the order-by item input to match to those.

    possible_inputs is necessary because at parse-time for string inputs, order-by inputs are resolved to spec patterns,
    and those patterns could match to either metrics or group-by-items.
    """

    input_obj: Union[str, OrderByQueryParameter]
    possible_inputs: Tuple[Union[ResolverInputForMetric, ResolverInputForGroupByItem], ...]
    descending: bool

    @property
    @override
    def ui_description(self) -> str:
        return str(self.input_obj)


@dataclass(frozen=True)
class ResolverInputForLimit(MetricFlowQueryResolverInput):
    """An input that describes the limit."""

    limit: Optional[int]

    @property
    @override
    def ui_description(self) -> str:
        return str(self.limit)


@dataclass(frozen=True)
class ResolverInputForMinMaxOnly(MetricFlowQueryResolverInput):
    """An input that describes if the query will be aggregated to only the min and max."""

    min_max_only: bool = False

    @property
    @override
    def ui_description(self) -> str:
        return str(self.min_max_only)


@dataclass(frozen=True)
class ResolverInputForQueryLevelWhereFilterIntersection(MetricFlowQueryResolverInput):
    """An input that describes the where filter for the query."""

    where_filter_intersection: WhereFilterIntersection

    @property
    @override
    def ui_description(self) -> str:
        return (
            "WhereFilter(\n"
            + indent(
                mf_pformat(
                    [where_filter.where_sql_template for where_filter in self.where_filter_intersection.where_filters]
                )
            )
            + "\n)"
        )


@dataclass(frozen=True)
class ResolverInputForWhereFilterIntersection(MetricFlowQueryResolverInput):
    """An input to describe a where filter anywhere in the resolution DAG.

    This can represent a query-level where filter, or one that is in a metric definition. Treating the defined filters
    as a resolver input so that it can go through the same error flows as the other query inputs.
    """

    where_filter_intersection: WhereFilterIntersection
    filter_resolution_path: MetricFlowQueryResolutionPath
    # e.g. TimeDimension('metric_time'), but may not be available if filter has parsing errors.
    object_builder_str: Optional[str]

    @property
    @override
    def ui_description(self) -> str:
        lines = [
            "WhereFilter(",
            indent(
                mf_pformat(
                    [where_filter.where_sql_template for where_filter in self.where_filter_intersection.where_filters]
                )
            ),
            ")",
            "Filter Path:",
            indent(self.filter_resolution_path.ui_description),
        ]
        if self.object_builder_str is not None:
            lines.append("Object Builder Input:")
            lines.append(indent(self.object_builder_str))
        return "\n".join(lines)


@dataclass(frozen=True)
class ResolverInputForQuery(MetricFlowQueryResolverInput):
    """An input that describes the entire query."""

    metric_inputs: Tuple[ResolverInputForMetric, ...]
    group_by_item_inputs: Tuple[ResolverInputForGroupByItem, ...]
    filter_input: ResolverInputForQueryLevelWhereFilterIntersection
    order_by_item_inputs: Tuple[ResolverInputForOrderByItem, ...]
    limit_input: ResolverInputForLimit
    min_max_only: ResolverInputForMinMaxOnly

    @property
    @override
    def ui_description(self) -> str:
        return (
            f"Query({repr([metric_input.ui_description for metric_input in self.metric_inputs])}, "
            f"{repr([group_by_item_input.input_obj for group_by_item_input in self.group_by_item_inputs])}"
        )



from __future__ import annotations

from typing import Collection

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from metricflow_semantics.errors.error_classes import UnsupportedEngineFeatureError
from metricflow_semantics.sql.sql_exprs import SqlPercentileExpression, SqlPercentileFunctionType
from typing_extensions import override

from metricflow.protocols.sql_client import SqlEngine
from metricflow.sql.render.expr_renderer import (
    DefaultSqlExpressionRenderer,
    SqlExpressionRenderer,
    SqlExpressionRenderResult,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlPlanRenderer


class DatabricksSqlExpressionRenderer(DefaultSqlExpressionRenderer):
    """Expression renderer for the Databricks engine."""

    sql_engine = SqlEngine.DATABRICKS

    @property
    @override
    def supported_percentile_function_types(self) -> Collection[SqlPercentileFunctionType]:
        return {SqlPercentileFunctionType.CONTINUOUS, SqlPercentileFunctionType.APPROXIMATE_DISCRETE}

    @override
    def render_date_part(self, date_part: DatePart) -> str:
        if date_part is DatePart.DOW:
            return "DAYOFWEEK_ISO"

        return super().render_date_part(date_part)

    @override
    def visit_percentile_expr(self, node: SqlPercentileExpression) -> SqlExpressionRenderResult:
        """Render a percentile expression for Databricks."""
        arg_rendered = self.render_sql_expr(node.order_by_arg)
        params = arg_rendered.bind_parameter_set
        percentile = node.percentile_args.percentile

        if node.percentile_args.function_type is SqlPercentileFunctionType.CONTINUOUS:
            function_str = "PERCENTILE_CONT"
        elif node.percentile_args.function_type is SqlPercentileFunctionType.DISCRETE:
            # discrete percentile only supported on databricks 11.0 >=. Disable for now.
            raise UnsupportedEngineFeatureError(
                "Discrete percentile aggregate not supported for Databricks.  Use "
                + "continuous or approximate discrete percentile in all percentile measures."
            )
        elif node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_CONTINUOUS:
            raise UnsupportedEngineFeatureError(
                "Approximate continuous percentile aggregate not supported for Databricks. Use "
                + "continuous or approximate discrete percentile in all percentile measures."
            )
        elif node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_DISCRETE:
            return SqlExpressionRenderResult(
                sql=f"APPROX_PERCENTILE({arg_rendered.sql}, {percentile})",
                bind_parameter_set=params,
            )
        else:
            assert_values_exhausted(node.percentile_args.function_type)

        return SqlExpressionRenderResult(
            sql=f"{function_str}({percentile}) WITHIN GROUP (ORDER BY ({arg_rendered.sql}))",
            bind_parameter_set=params,
        )


class DatabricksSqlPlanRenderer(DefaultSqlPlanRenderer):
    """Plan renderer for the Snowflake engine."""

    EXPR_RENDERER = DatabricksSqlExpressionRenderer()

    @property
    @override
    def expr_renderer(self) -> SqlExpressionRenderer:
        return self.EXPR_RENDERER



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Sequence, Set

from typing_extensions import override

from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.spec_set import group_specs_by_type
from metricflow_semantics.specs.time_dimension_spec import (
    TimeDimensionSpec,
    TimeDimensionSpecComparisonKey,
    TimeDimensionSpecField,
)
from metricflow_semantics.time.granularity import ExpandedTimeGranularity


@dataclass(frozen=True)
class MinimumTimeGrainPattern(SpecPattern):
    """A pattern that matches linkable specs, but for time dimension specs, only the one with the finest base grain.

    e.g.

    inputs:
        [
            TimeDimensionSpec('metric_time', 'day'),
            TimeDimensionSpec('metric_time', 'month.'),
            DimensionSpec('listing__country'),
        ]

    matches:
        [
            TimeDimensionSpec('metric_time', 'day'),
            DimensionSpec('listing__country'),
        ]

    The finest grain represents the defined grain of the time dimension in the semantic model when evaluating specs
    of the source. For custom granularities, this means the base grain associated with the time dimension spec.

    This pattern helps to implement matching of group-by-items for where filters - in those cases, an ambiguously
    specified group-by-item can only match to time dimension spec with the base grain.
    """

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[InstanceSpec]:
        spec_set = group_specs_by_type(candidate_specs)

        spec_key_to_grains: Dict[TimeDimensionSpecComparisonKey, Set[ExpandedTimeGranularity]] = defaultdict(set)
        spec_key_to_specs: Dict[TimeDimensionSpecComparisonKey, List[TimeDimensionSpec]] = defaultdict(list)
        for time_dimension_spec in spec_set.time_dimension_specs:
            spec_key = time_dimension_spec.comparison_key(exclude_fields=(TimeDimensionSpecField.TIME_GRANULARITY,))
            spec_key_to_grains[spec_key].add(time_dimension_spec.time_granularity)
            spec_key_to_specs[spec_key].append(time_dimension_spec)

        matched_time_dimension_specs: List[TimeDimensionSpec] = []
        for spec_key, time_grains in spec_key_to_grains.items():
            sorted_time_grains = sorted(
                time_grains,
                # Sort by smallest to largest standard granularity, with custom granularities last (sorted by their
                # base granularity) since we don't know how large they are.
                key=lambda grain: (grain.is_custom_granularity, grain.base_granularity.to_int()),
            )
            assert sorted_time_grains, (
                f"Each time dimension spec should have at least one grain, but none was found for spec_key {spec_key}. "
                "This indicates internal misconfiguration."
            )
            matched_time_dimension_specs.append(spec_key_to_specs[spec_key][0].with_grain(sorted_time_grains[0]))

        matching_specs: Sequence[LinkableInstanceSpec] = (
            spec_set.dimension_specs
            + tuple(matched_time_dimension_specs)
            + spec_set.entity_specs
            + spec_set.group_by_metric_specs
        )

        return matching_specs



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Mapping, Sequence, Tuple

from dbt_semantic_interfaces.references import TimeDimensionReference
from dbt_semantic_interfaces.type_enums import TimeGranularity
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.query.query_parser import MetricFlowQueryParser
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
from metricflow_semantics.specs.query_spec import MetricFlowQuerySpec
from metricflow_semantics.time.time_spine_source import TimeSpineSource

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
from metricflow.dataflow.nodes.read_sql_source import ReadSqlSourceNode
from metricflow.dataset.convert_semantic_model import SemanticModelToDataSetConverter
from metricflow.dataset.semantic_model_adapter import SemanticModelDataSet


@dataclass(frozen=True)
class SourceNodeSet:
    """Contains nodes / components that are used by the `DataflowPlanBuilder` as root building blocks.

    The components in this set do not need to be dynamically generated on a per-query basis for a given semantic
    manifest.
    """

    # Semantic models without measures are 1:1 mapped to a ReadSqlSourceNode. Semantic models containing measures are
    # mapped to components with a transformation node to add `metric_time` / to support multiple aggregation time
    # dimensions. Each semantic model containing measures with k different aggregation time dimensions is mapped to k
    # components.
    source_nodes_for_metric_queries: Tuple[DataflowPlanNode, ...]

    # Semantic models are 1:1 mapped to a ReadSqlSourceNode.
    source_nodes_for_group_by_item_queries: Tuple[DataflowPlanNode, ...]

    # Provides time spines that can be used to satisfy time spine joins.
    time_spine_read_nodes: Mapping[TimeGranularity, ReadSqlSourceNode]

    # Provides time spines that can be used to satisfy metric_time without metrics.
    time_spine_metric_time_nodes: Mapping[TimeGranularity, MetricTimeDimensionTransformNode]

    @property
    def all_nodes(self) -> Sequence[DataflowPlanNode]:  # noqa: D102
        return (
            self.source_nodes_for_metric_queries
            + self.source_nodes_for_group_by_item_queries
            + self.time_spine_metric_time_nodes_tuple
        )

    @property
    def time_spine_metric_time_nodes_tuple(self) -> Tuple[MetricTimeDimensionTransformNode, ...]:  # noqa: D102
        return tuple(self.time_spine_metric_time_nodes.values())


class SourceNodeBuilder:
    """Helps build a `SourceNodeSet` - refer to that class for more details."""

    def __init__(  # noqa: D107
        self,
        column_association_resolver: ColumnAssociationResolver,
        semantic_manifest_lookup: SemanticManifestLookup,
    ) -> None:
        self._semantic_manifest_lookup = semantic_manifest_lookup
        data_set_converter = SemanticModelToDataSetConverter(column_association_resolver)
        self.time_spine_sources = TimeSpineSource.build_standard_time_spine_sources(
            semantic_manifest_lookup.semantic_manifest
        )

        self._time_spine_read_nodes = {}
        self._time_spine_metric_time_nodes = {}
        for base_granularity, time_spine_source in self.time_spine_sources.items():
            data_set = data_set_converter.build_time_spine_source_data_set(time_spine_source)
            read_node = ReadSqlSourceNode.create(data_set)
            self._time_spine_read_nodes[base_granularity] = read_node
            self._time_spine_metric_time_nodes[base_granularity] = MetricTimeDimensionTransformNode.create(
                parent_node=read_node,
                aggregation_time_dimension_reference=TimeDimensionReference(time_spine_source.base_column),
            )

        self._query_parser = MetricFlowQueryParser(semantic_manifest_lookup)

    def create_from_data_sets(self, data_sets: Sequence[SemanticModelDataSet]) -> SourceNodeSet:
        """Creates a `SourceNodeSet` from SemanticModelDataSets."""
        group_by_item_source_nodes: List[DataflowPlanNode] = []
        source_nodes_for_metric_queries: List[DataflowPlanNode] = []

        for data_set in data_sets:
            read_node = ReadSqlSourceNode.create(data_set)
            group_by_item_source_nodes.append(read_node)
            agg_time_dim_to_measures_grouper = (
                self._semantic_manifest_lookup.semantic_model_lookup.get_aggregation_time_dimensions_with_measures(
                    data_set.semantic_model_reference
                )
            )

            # Dimension sources may not have any measures -> no aggregation time dimensions.
            time_dimension_references = agg_time_dim_to_measures_grouper.keys
            if len(time_dimension_references) == 0:
                source_nodes_for_metric_queries.append(read_node)
            else:
                # Splits the measures by distinct aggregate time dimension.
                for time_dimension_reference in time_dimension_references:
                    metric_time_transform_node = MetricTimeDimensionTransformNode.create(
                        parent_node=read_node,
                        aggregation_time_dimension_reference=time_dimension_reference,
                    )
                    source_nodes_for_metric_queries.append(metric_time_transform_node)

        return SourceNodeSet(
            time_spine_metric_time_nodes=self._time_spine_metric_time_nodes,
            time_spine_read_nodes=self._time_spine_read_nodes,
            source_nodes_for_group_by_item_queries=tuple(group_by_item_source_nodes),
            source_nodes_for_metric_queries=tuple(source_nodes_for_metric_queries),
        )

    def build_source_node_inputs_for_group_by_metric(
        self, group_by_metric_spec: GroupByMetricSpec
    ) -> MetricFlowQuerySpec:
        """Build source node inputs used to satisfy requested group by metrics.

        Group by metrics are essentially metric subqueries that operate as source nodes in the DataflowPlanBuilder. We
        provide the inputs here because they require an entire DataFlowPlan generation step.

        This is just a wrapper around the query parser method, stored here to limit the scope of the DataFlowPlanBuilder's
        dependency on the query parser to only source nodes.
        """
        return self._query_parser.build_query_spec_for_group_by_metric_source_node(group_by_metric_spec)



from __future__ import annotations

import logging
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DagId, DagNode, DisplayedProperty, MetricFlowDag, NodeId
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.visitor import Visitable

from metricflow.data_table.mf_table import MetricFlowDataTable
from metricflow.protocols.sql_client import SqlClient

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ExecutionPlanTask(DagNode["ExecutionPlanTask"], Visitable, ABC):
    """A node (aka task) in the DAG representation of the execution plan.

    In the DAG, a node's parents represent the tasks that need to be run before the node can run. Using the term task
    for these nodes as it seems more intuitive.

    Attributes:
        sql_statement: If this runs a SQL query, return the associated SQL.
    """

    sql_statement: Optional[SqlStatement]

    @abstractmethod
    def execute(self) -> TaskExecutionResult:
        """Execute the actions of this node."""
        raise NotImplementedError

    @property
    def task_id(self) -> NodeId:
        """Alias for node ID since the nodes represent a task."""
        return self.node_id


@dataclass(frozen=True)
class SqlStatement:
    """Encapsulates a SQL statement along with the bind parameters that should be used."""

    # This field will be renamed as it is confusing given the class name.
    sql: str
    bind_parameter_set: SqlBindParameterSet

    @property
    def without_descriptions(self) -> SqlStatement:
        """Return the SQL query without the inline descriptions."""
        return SqlStatement(
            sql="\n".join(
                filter(
                    lambda line: not line.strip().startswith("--"),
                    self.sql.split("\n"),
                )
            ),
            bind_parameter_set=self.bind_parameter_set,
        )


@dataclass(frozen=True)
class TaskExecutionError(Exception):
    """Error if a task fails."""

    error_str: str


@dataclass(frozen=True)
class TaskExecutionResult:
    """The results of running a task."""

    start_time: float
    end_time: float
    errors: Tuple[TaskExecutionError, ...] = ()

    # If the task was an SQL query, it's stored here
    sql: Optional[str] = None
    bind_params: Optional[SqlBindParameterSet] = None
    # If the task produces a data_table as a result, it's stored here.
    df: Optional[MetricFlowDataTable] = None


@dataclass(frozen=True)
class SelectSqlQueryToDataTableTask(ExecutionPlanTask):
    """A task that runs a SELECT and puts that result into a data_table.

    Attributes:
        sql_client: The SQL client used to run the query.
        sql_statement: The SQL query to run.
        parent_nodes: The parent tasks for this execution plan task.
    """

    sql_client: SqlClient
    parent_nodes: Tuple[ExecutionPlanTask, ...]

    @staticmethod
    def create(  # noqa: D102
        sql_client: SqlClient,
        sql_statement: SqlStatement,
        parent_nodes: Sequence[ExecutionPlanTask] = (),
    ) -> SelectSqlQueryToDataTableTask:
        return SelectSqlQueryToDataTableTask(
            sql_client=sql_client,
            sql_statement=sql_statement,
            parent_nodes=tuple(parent_nodes),
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.EXEC_NODE_READ_SQL_QUERY

    @property
    def description(self) -> str:  # noqa: D102
        return "Run a query and write the results to a data frame"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        assert self.sql_statement is not None, f"{self.sql_statement=} should have been set during creation."
        return tuple(super().displayed_properties) + (DisplayedProperty(key="sql", value=self.sql_statement.sql),)

    def execute(self) -> TaskExecutionResult:  # noqa: D102
        start_time = time.time()
        sql_statement = self.sql_statement
        assert sql_statement is not None, f"{self.sql_statement=} should have been set during creation."

        df = self.sql_client.query(
            sql_statement.sql,
            sql_bind_parameter_set=sql_statement.bind_parameter_set,
        )

        end_time = time.time()
        return TaskExecutionResult(
            start_time=start_time,
            end_time=end_time,
            sql=sql_statement.sql,
            bind_params=sql_statement.bind_parameter_set,
            df=df,
        )

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(sql_statement={self.sql_statement!r})"


@dataclass(frozen=True)
class SelectSqlQueryToTableTask(ExecutionPlanTask):
    """A task that runs a SELECT and puts that result into a table.

    The provided SQL query is the query that will be run, so it should be a CREATE... or similar.

    Attributes:
        sql_client: The SQL client used to run the query.
        sql_statement: The SQL query to run.
        output_table: The table where the results will be written.
    """

    sql_client: SqlClient
    output_table: SqlTable

    @staticmethod
    def create(  # noqa: D102
        sql_client: SqlClient,
        sql_statement: SqlStatement,
        output_table: SqlTable,
        parent_nodes: Sequence[ExecutionPlanTask] = (),
    ) -> SelectSqlQueryToTableTask:
        return SelectSqlQueryToTableTask(
            sql_client=sql_client,
            sql_statement=sql_statement,
            output_table=output_table,
            parent_nodes=tuple(parent_nodes),
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.EXEC_NODE_WRITE_TO_TABLE

    @property
    def description(self) -> str:  # noqa: D102
        return "Run a query and write the results to a table"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        sql_statement = self.sql_statement
        assert sql_statement is not None, f"{self.sql_statement=} should have been set during creation."
        return tuple(super().displayed_properties) + (
            DisplayedProperty(key="sql_statement", value=sql_statement.sql),
            DisplayedProperty(key="output_table", value=self.output_table),
            DisplayedProperty(key="bind_parameter_set", value=sql_statement.bind_parameter_set),
        )

    def execute(self) -> TaskExecutionResult:  # noqa: D102
        sql_statement = self.sql_statement
        assert sql_statement is not None, f"{self.sql_statement=} should have been set during creation."
        start_time = time.time()
        logger.debug(LazyFormat(lambda: f"Dropping table {self.output_table} in case it already exists"))
        self.sql_client.execute(f"DROP TABLE IF EXISTS {self.output_table.sql}")
        logger.debug(LazyFormat(lambda: f"Creating table {self.output_table} using a query"))
        self.sql_client.execute(
            sql_statement.sql,
            sql_bind_parameter_set=sql_statement.bind_parameter_set,
        )

        end_time = time.time()
        return TaskExecutionResult(start_time=start_time, end_time=end_time, sql=sql_statement.sql)

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(sql_statement={self.sql_statement!r}', output_table={self.output_table})"


class ExecutionPlan(MetricFlowDag[ExecutionPlanTask]):
    """A DAG where the nodes are tasks, and parents represent prerequisite tasks."""

    def __init__(self, leaf_tasks: Sequence[ExecutionPlanTask], dag_id: Optional[DagId] = None) -> None:
        """Constructor.

        Args:
            leaf_tasks: The final set of tasks that will run, after task dependencies are finished.
        """
        super().__init__(
            dag_id=dag_id or DagId.from_id_prefix(StaticIdPrefix.EXEC_PLAN_PREFIX), sink_nodes=tuple(leaf_tasks)
        )

    @property
    def tasks(self) -> Sequence[ExecutionPlanTask]:
        """Return all tasks in this plan."""
        if len(self.sink_nodes) == 0:
            return ()

        def recursively_get_tasks(task: ExecutionPlanTask) -> List[ExecutionPlanTask]:
            tasks_to_return = []
            for parent_node in task.parent_nodes:
                tasks_to_return.extend(recursively_get_tasks(parent_node))
            tasks_to_return.append(task)

            return tasks_to_return

        assert len(self.sink_nodes) == 1
        return tuple(recursively_get_tasks(self.sink_nodes[0]))



from __future__ import annotations

import logging
from typing import Optional, Sequence, Tuple

from dbt_semantic_interfaces.type_enums.date_part import DatePart
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow_semantics.collection_helpers.lru_cache import typed_lru_cache

DUNDER = "__"

logger = logging.getLogger(__name__)


class StructuredLinkableSpecName:
    """Parse a qualified name into different parts.

    e.g. listing__ds__week ->
    entity_links: ["listing"]
    element_name: "ds"
    granularity: TimeGranularity.WEEK
    """

    def __init__(
        self,
        entity_link_names: Tuple[str, ...],
        element_name: str,
        time_granularity_name: Optional[str] = None,
        date_part: Optional[DatePart] = None,
    ) -> None:
        """Set attributes, ensuring names are lowercased."""
        self.entity_link_names = tuple([entity_link_name.lower() for entity_link_name in entity_link_names])
        self.element_name = element_name.lower()
        self.time_granularity_name = time_granularity_name.lower() if time_granularity_name else None
        self.date_part = date_part

    @staticmethod
    @typed_lru_cache
    def from_name(qualified_name: str, custom_granularity_names: Sequence[str]) -> StructuredLinkableSpecName:
        """Construct from a name e.g. listing__ds__month."""
        name_parts = qualified_name.split(DUNDER)

        # No dunder, e.g. "ds"
        if len(name_parts) == 1:
            return StructuredLinkableSpecName(entity_link_names=(), element_name=name_parts[0])

        for date_part in DatePart:
            if name_parts[-1] == StructuredLinkableSpecName.date_part_suffix(date_part=date_part):
                raise ValueError(
                    "Dunder syntax not supported for querying date_part. Use `group_by` object syntax instead."
                )

        associated_granularity: Optional[str] = None
        for granularity in TimeGranularity:
            if name_parts[-1] == granularity.value:
                associated_granularity = granularity.value
                break

        if associated_granularity is None:
            for custom_grain in custom_granularity_names:
                if name_parts[-1] == custom_grain:
                    associated_granularity = custom_grain
                    break

        # Has a time granularity
        if associated_granularity:
            #  e.g. "ds__month"
            if len(name_parts) == 2:
                return StructuredLinkableSpecName(
                    entity_link_names=(), element_name=name_parts[0], time_granularity_name=associated_granularity
                )
            # e.g. "messages__ds__month"
            return StructuredLinkableSpecName(
                entity_link_names=tuple(name_parts[:-2]),
                element_name=name_parts[-2],
                time_granularity_name=associated_granularity,
            )

        # e.g. "messages__ds"
        else:
            return StructuredLinkableSpecName(entity_link_names=tuple(name_parts[:-1]), element_name=name_parts[-1])

    @property
    def qualified_name(self) -> str:
        """Return the full name form. e.g. ds or listing__ds__month.

        If date_part is specified, don't include granularity in qualified_name since it will not impact the result.
        """
        items = list(self.entity_link_names) + [self.element_name]
        if self.date_part:
            items.append(self.date_part_suffix(date_part=self.date_part))
        elif self.time_granularity_name:
            items.append(self.time_granularity_name)
        return DUNDER.join(items)

    @property
    def entity_prefix(self) -> Optional[str]:
        """Return the entity prefix. e.g. listing__ds__month -> listing."""
        if len(self.entity_link_names) > 0:
            return DUNDER.join(self.entity_link_names)

        return None

    @staticmethod
    def date_part_suffix(date_part: DatePart) -> str:
        """Suffix used for names with a date_part."""
        return f"extract_{date_part.value}"

    @property
    def granularity_free_qualified_name(self) -> str:
        """Renders the qualified name without the granularity suffix.

        In the list metrics and list dimensions outputs we want to render the qualified name of the dimension, but
        without including the base granularity for time dimensions. This method is useful in those contexts.
        Note: in most cases you should be using the qualified_name - this is only useful in cases where the
        Dimension set has de-duplicated TimeDimensions such that you never have more than one granularity
        in your set for each TimeDimension.
        """
        return StructuredLinkableSpecName(
            entity_link_names=self.entity_link_names, element_name=self.element_name
        ).qualified_name

    @property
    def is_element_name(self) -> bool:
        """Indicates whether or not this is an unadorned element name, i.e., one without links or time annotations."""
        return self.qualified_name == self.element_name



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.protocols.metric import MetricTimeWindow
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from metricflow_semantics.assert_one_arg import assert_exactly_one_arg_set
from metricflow_semantics.sql.sql_exprs import (
    SqlColumnReference,
    SqlColumnReferenceExpression,
    SqlComparison,
    SqlComparisonExpression,
    SqlDateTruncExpression,
    SqlIsNullExpression,
    SqlLogicalExpression,
    SqlLogicalOperator,
    SqlSubtractTimeIntervalExpression,
)
from metricflow_semantics.sql.sql_join_type import SqlJoinType

from metricflow.dataflow.nodes.join_conversion_events import JoinConversionEventsNode
from metricflow.dataflow.nodes.join_over_time import JoinOverTimeRangeNode
from metricflow.dataflow.nodes.join_to_base import JoinDescription
from metricflow.dataflow.nodes.join_to_time_spine import JoinToTimeSpineNode
from metricflow.dataset.sql_dataset import AnnotatedSqlDataSet
from metricflow.plan_conversion.sql_expression_builders import make_coalesced_expr
from metricflow.sql.sql_plan import SqlExpressionNode, SqlJoinDescription, SqlSelectStatementNode


@dataclass(frozen=True)
class ColumnEqualityDescription:
    """Helper class to enumerate columns that should be equal between sources in a join.

    The `treat_nulls_as_equal` property determines what to do with null valued inputs. By default SQL returns
    NULL for any equality comparison with a NULL on either side, which means `NULL = NULL` returns NULL, which
    evaluates to False in the join comparison. If that parameter is overridden to True, then the column comparison
    will be rendered as:

        `(left_column_alias = right_column_alias OR (left_column_alias IS NULL AND right_column_alias IS NULL))`
    """

    left_column_alias: str
    right_column_alias: str
    treat_nulls_as_equal: bool = False


class SqlPlanJoinBuilder:
    """Helper class for constructing various join components in a SqlQueryPlan."""

    @staticmethod
    def make_column_equality_sql_join_description(
        right_source_node: SqlSelectStatementNode,
        left_source_alias: str,
        right_source_alias: str,
        column_equality_descriptions: Sequence[ColumnEqualityDescription],
        join_type: SqlJoinType,
        additional_on_conditions: Sequence[SqlExpressionNode] = tuple(),
    ) -> SqlJoinDescription:
        """Make a join description where the base condition is a set of equality comparisons between columns.

        Typically the columns in column_equality_descriptions are entities we are trying to match,
        although they may include things like dimension partitions or time dimension columns where an
        equality is expected.

        Since the framework currently supports join keys OR condition-free cross joins, this helper will require
        either a non-empty set of column equality conditions or a CROSS JOIN.

        Args:
            right_source_node: node representing the join target, may be either a table or subquery
            left_source_alias: string alias entity for the join source
            right_source_alias: string alias entity for the join target
            column_equality_descriptions: set of equality constraints for the ON statement
            join_type: type of SQL join, e.g., LEFT, INNER, etc.
            additional_on_conditions: set of additional constraints to add in the ON statement (via AND)
        """
        assert (
            len(column_equality_descriptions) > 0 or join_type is SqlJoinType.CROSS_JOIN
        ), f"No column equality conditions specified for join with type {join_type} - this may render invalid SQL!"

        and_conditions: List[SqlExpressionNode] = []
        for column_equality_description in column_equality_descriptions:
            left_column = SqlColumnReferenceExpression.create(
                SqlColumnReference(
                    table_alias=left_source_alias,
                    column_name=column_equality_description.left_column_alias,
                )
            )
            right_column = SqlColumnReferenceExpression.create(
                SqlColumnReference(
                    table_alias=right_source_alias,
                    column_name=column_equality_description.right_column_alias,
                )
            )
            column_equality_expression = SqlComparisonExpression.create(
                left_expr=left_column,
                comparison=SqlComparison.EQUALS,
                right_expr=right_column,
            )
            if column_equality_description.treat_nulls_as_equal:
                null_comparison_expression = SqlLogicalExpression.create(
                    operator=SqlLogicalOperator.AND,
                    args=(SqlIsNullExpression.create(arg=left_column), SqlIsNullExpression.create(arg=right_column)),
                )
                and_conditions.append(
                    SqlLogicalExpression.create(
                        operator=SqlLogicalOperator.OR, args=(column_equality_expression, null_comparison_expression)
                    )
                )
            else:
                and_conditions.append(column_equality_expression)

        and_conditions += additional_on_conditions

        on_condition: Optional[SqlExpressionNode]
        if len(and_conditions) == 0:
            on_condition = None
        elif len(and_conditions) == 1:
            on_condition = and_conditions[0]
        else:
            on_condition = SqlLogicalExpression.create(operator=SqlLogicalOperator.AND, args=tuple(and_conditions))

        return SqlJoinDescription(
            right_source=right_source_node,
            right_source_alias=right_source_alias,
            on_condition=on_condition,
            join_type=join_type,
        )

    @staticmethod
    def make_base_output_join_description(
        left_data_set: AnnotatedSqlDataSet,
        right_data_set: AnnotatedSqlDataSet,
        join_description: JoinDescription,
    ) -> SqlJoinDescription:
        """Make a join description to link two base output DataSets by matching entities.

        In addition to the entity equality condition, this will ensure datasets are joined on all partition
        columns and account for validity windows, if those are defined in one of the datasets.
        """
        join_on_entity = join_description.join_on_entity

        column_equality_descriptions = []
        if join_on_entity:
            # Figure out which columns in the "left" data set correspond to the entity that we want to join on.
            # The column associations tell us which columns correspond to which instances in the data set.
            left_data_set_entity_column_associations = left_data_set.data_set.column_associations_for_entity(
                join_on_entity
            )
            left_data_set_entity_cols = [c.column_name for c in left_data_set_entity_column_associations]

            # Figure out which columns in the "right" data set correspond to the entity that we want to join on.
            right_data_set_column_associations = right_data_set.data_set.column_associations_for_entity(join_on_entity)
            right_data_set_entity_cols = [c.column_name for c in right_data_set_column_associations]

            assert len(left_data_set_entity_cols) == len(right_data_set_entity_cols), (
                f"Cannot construct join - the number of columns on the left ({left_data_set_entity_cols}) side of "
                f"the join does not match the right ({right_data_set_entity_cols})"
            )

            # We have the columns that we need to "join on" in the query, so add it to the list of join descriptions to
            # use later.
            for idx in range(len(left_data_set_entity_cols)):
                column_equality_descriptions.append(
                    ColumnEqualityDescription(
                        left_column_alias=left_data_set_entity_cols[idx],
                        right_column_alias=right_data_set_entity_cols[idx],
                    )
                )

        # Add the partition columns as well.
        for dimension_join_description in join_description.join_on_partition_dimensions:
            column_equality_descriptions.append(
                ColumnEqualityDescription(
                    left_column_alias=left_data_set.data_set.column_association_for_dimension(
                        dimension_join_description.start_node_dimension_spec
                    ).column_name,
                    right_column_alias=right_data_set.data_set.column_association_for_dimension(
                        dimension_join_description.node_to_join_dimension_spec
                    ).column_name,
                )
            )

        for time_dimension_join_description in join_description.join_on_partition_time_dimensions:
            column_equality_descriptions.append(
                ColumnEqualityDescription(
                    left_column_alias=left_data_set.data_set.column_association_for_time_dimension(
                        time_dimension_join_description.start_node_time_dimension_spec
                    ).column_name,
                    right_column_alias=right_data_set.data_set.column_association_for_time_dimension(
                        time_dimension_join_description.node_to_join_time_dimension_spec
                    ).column_name,
                )
            )

        validity_conditions = SqlPlanJoinBuilder._make_validity_window_on_conditions(
            left_data_set=left_data_set, right_data_set=right_data_set, join_description=join_description
        )

        return SqlPlanJoinBuilder.make_column_equality_sql_join_description(
            right_source_node=right_data_set.data_set.checked_sql_select_node,
            left_source_alias=left_data_set.alias,
            right_source_alias=right_data_set.alias,
            column_equality_descriptions=column_equality_descriptions,
            join_type=join_description.join_type,
            additional_on_conditions=validity_conditions,
        )

    @staticmethod
    def _make_validity_window_on_conditions(
        left_data_set: AnnotatedSqlDataSet,
        right_data_set: AnnotatedSqlDataSet,
        join_description: JoinDescription,
    ) -> Tuple[SqlExpressionNode, ...]:
        """Build a time window join condition if the join description includes a validity window description.

        When the validity window is set, it means we are dealing with a dataset representing an SCD Type II
        style semantic model, with a start and end boundary on the window. A base output join against such data
        sources requires a metric_time dimension to process the join.

        By convention, the join description ties the validity window to the "right" side semantic model, and so
        we extract the metric time instance from the left data set.

        We use the instance with the smallest granularity and shortest entity link path, since it will be
        used in the ON statement for the join against the validity window.

        Note SCD joins are always evaluated against a standard time granularity. Subsequent conversion to
        custom granularities should happen after the join.
        """
        if join_description.validity_window is None:
            return tuple()

        left_data_set_metric_time_dimension_instances = sorted(
            [
                instance
                for instance in left_data_set.data_set.metric_time_dimension_instances
                if not instance.spec.time_granularity.is_custom_granularity
            ],
            key=lambda x: (x.spec.time_granularity.base_granularity.to_int(), len(x.spec.entity_links)),
        )
        assert left_data_set_metric_time_dimension_instances, (
            f"Cannot process join to data set with alias {right_data_set.alias} because it has a validity "
            f"window set: {join_description.validity_window}, but source data set with alias "
            f"{left_data_set.alias} does not have a metric time dimension we can use for the window join!"
        )
        left_data_set_time_dimension_name = left_data_set.data_set.column_association_for_time_dimension(
            left_data_set_metric_time_dimension_instances[0].spec,
        ).column_name
        window_start_dimension_name = right_data_set.data_set.column_association_for_time_dimension(
            join_description.validity_window.window_start_dimension
        ).column_name
        window_end_dimension_name = right_data_set.data_set.column_association_for_time_dimension(
            join_description.validity_window.window_end_dimension
        ).column_name
        window_join_condition = SqlPlanJoinBuilder._make_time_window_join_condition(
            left_source_alias=left_data_set.alias,
            left_source_time_dimension_name=left_data_set_time_dimension_name,
            right_source_alias=right_data_set.alias,
            window_start_dimension_name=window_start_dimension_name,
            window_end_dimension_name=window_end_dimension_name,
        )

        return (window_join_condition,)

    @staticmethod
    def _make_time_window_join_condition(
        left_source_alias: str,
        left_source_time_dimension_name: str,
        right_source_alias: str,
        window_start_dimension_name: str,
        window_end_dimension_name: str,
    ) -> SqlLogicalExpression:
        """Helper to generate a renderable SqlExpression for expressing a time window join condition.

        The output will render the following expression:

            {start_dimension_name} >= metric_time AND ({end_dimension_name} < metric_time OR {end_dimension_name} IS NULL)

        """
        left_time_column_expr = SqlColumnReferenceExpression.create(
            SqlColumnReference(table_alias=left_source_alias, column_name=left_source_time_dimension_name)
        )
        window_start_column_expr = SqlColumnReferenceExpression.create(
            SqlColumnReference(table_alias=right_source_alias, column_name=window_start_dimension_name)
        )
        window_end_column_expr = SqlColumnReferenceExpression.create(
            SqlColumnReference(table_alias=right_source_alias, column_name=window_end_dimension_name)
        )

        window_start_condition = SqlComparisonExpression.create(
            left_expr=left_time_column_expr,
            comparison=SqlComparison.GREATER_THAN_OR_EQUALS,
            right_expr=window_start_column_expr,
        )
        window_end_by_time = SqlComparisonExpression.create(
            left_expr=left_time_column_expr,
            comparison=SqlComparison.LESS_THAN,
            right_expr=window_end_column_expr,
        )
        window_end_is_null = SqlIsNullExpression.create(window_end_column_expr)
        window_end_condition = SqlLogicalExpression.create(
            operator=SqlLogicalOperator.OR, args=(window_end_by_time, window_end_is_null)
        )

        return SqlLogicalExpression.create(
            operator=SqlLogicalOperator.AND, args=(window_start_condition, window_end_condition)
        )

    @staticmethod
    def make_join_description_for_combining_datasets(
        from_data_set: AnnotatedSqlDataSet,
        join_data_set: AnnotatedSqlDataSet,
        join_type: SqlJoinType,
        column_names: Sequence[str],
        table_aliases_for_coalesce: Sequence[str],
    ) -> SqlJoinDescription:
        """Creates the sql join description for combining two separate output datasets.

        These might be combined in service of producing complete output for end user consumption, in which
        case the join type will be FULL OUTER in order to ensure all rows are included. In this case, the
        join must account for all past table aliases seen in prior joins in order to produce the final
        coalesce expression for processing the join correctly.

        Metric data sets might also be combined to serve as inputs to a derived metric, in which case the
        join type will not be a FULL OUTER, and we should use null-safe column equality comparisons with
        the relevant join type for
        """
        if join_type is SqlJoinType.FULL_OUTER:
            assert (
                len(column_names) > 0
            ), "Attempting to do a FULL OUTER JOIN to combine metrics, but no columns were provided for join keys!"
            equality_exprs = [
                SqlPlanJoinBuilder._make_equality_expression_for_full_outer_join(
                    table_aliases_for_coalesce, join_data_set.alias, colname
                )
                for colname in column_names
            ]
            on_condition = (
                SqlLogicalExpression.create(operator=SqlLogicalOperator.AND, args=tuple(equality_exprs))
                if len(equality_exprs) > 1
                else equality_exprs[0]
            )
            return SqlJoinDescription(
                right_source=join_data_set.data_set.checked_sql_select_node,
                right_source_alias=join_data_set.alias,
                on_condition=on_condition,
                join_type=join_type,
            )
        else:
            column_equality_descriptions = [
                ColumnEqualityDescription(left_column_alias=name, right_column_alias=name, treat_nulls_as_equal=True)
                for name in column_names
            ]
            return SqlPlanJoinBuilder.make_column_equality_sql_join_description(
                right_source_node=join_data_set.data_set.checked_sql_select_node,
                left_source_alias=from_data_set.alias,
                right_source_alias=join_data_set.alias,
                column_equality_descriptions=column_equality_descriptions,
                join_type=join_type,
            )

    @staticmethod
    def _make_equality_expression_for_full_outer_join(
        table_aliases_in_coalesce: Sequence[str], right_table_alias: str, column_alias: str
    ) -> SqlExpressionNode:
        """Creates an expression that can go in the ON condition when coalescing join keys with a FULL OUTER JOIN.

        e.g.

        dimension_specs = ["is_instant", "ds"]
        table_aliases_in_coalesce = ["a", "b"]
        table_alias_on_right_equality = ["c"]

        ->

        COALESCE(a.is_instant, b.is_instant) = c.is_instant
        AND COALESCE(a.ds, b.ds) = c.ds

        This is necessary for cases where 3 or more subqueries will be linked via FULL OUTER JOINs. If that happens,
        the set of join keys from subquery 3 must be compared to the join keys from both subquery 1 and subquery 2,
        otherwise duplicate rows might appear in the output.

        For example:

        table1: [('a', 10), ('b', 20)]
        table2: [('a', 100), ('c', 200)]
        table3: [('a', 1000), ('c', 2000)]

        ->

        output without COALESCE: [('a', 10, 100, 1000), ('c', NULL, 200, NULL), ('c', NULL, NULL, 2000)]
        output with COALESCE: [('a', 10, 100, 1000), ('c', NULL, 200, 2000)]

        The latter scenario consolidates the rows keyed by 'c' into a single entry.
        """
        return SqlComparisonExpression.create(
            left_expr=make_coalesced_expr(table_aliases_in_coalesce, column_alias),
            comparison=SqlComparison.EQUALS,
            right_expr=SqlColumnReferenceExpression.create(
                col_ref=SqlColumnReference(
                    table_alias=right_table_alias,
                    column_name=column_alias,
                )
            ),
        )

    @staticmethod
    def _make_time_range_window_join_condition(
        base_data_set: AnnotatedSqlDataSet,
        time_comparison_dataset: AnnotatedSqlDataSet,
        window: Optional[MetricTimeWindow] = None,
        grain_to_date: Optional[TimeGranularity] = None,
    ) -> SqlLogicalExpression:
        """Helper to generate a renderable SqlExpression for expressing a time range window join condition.

        The output will render the following expression:

            a.ds <= b.ds AND a.ds > b.ds - <window>
            If no window is present we join across all time -> "a.ds <= b.ds"
        """
        if window or grain_to_date:
            assert_exactly_one_arg_set(window=window, grain_to_date=grain_to_date)
        base_column_expr = SqlColumnReferenceExpression.create(
            SqlColumnReference(
                table_alias=base_data_set.alias,
                column_name=base_data_set.metric_time_column_name,
            )
        )
        time_comparison_column_expr = SqlColumnReferenceExpression.create(
            SqlColumnReference(
                table_alias=time_comparison_dataset.alias,
                column_name=time_comparison_dataset.metric_time_column_name,
            )
        )

        # Comparison expression against the endpoint of the cumulative time range,
        # meaning the base metrc time must always be BEFORE the comparison metric time
        end_of_range_comparison_expression = SqlComparisonExpression.create(
            left_expr=base_column_expr,
            comparison=SqlComparison.LESS_THAN_OR_EQUALS,
            right_expr=time_comparison_column_expr,
        )

        comparison_expressions: List[SqlComparisonExpression] = [end_of_range_comparison_expression]
        if window:
            start_of_range_comparison_expr = SqlComparisonExpression.create(
                left_expr=base_column_expr,
                comparison=SqlComparison.GREATER_THAN,
                right_expr=SqlSubtractTimeIntervalExpression.create(
                    arg=time_comparison_column_expr,
                    count=window.count,
                    granularity=window.granularity,
                ),
            )
            comparison_expressions.append(start_of_range_comparison_expr)
        elif grain_to_date:
            start_of_range_comparison_expr = SqlComparisonExpression.create(
                left_expr=base_column_expr,
                comparison=SqlComparison.GREATER_THAN_OR_EQUALS,
                right_expr=SqlDateTruncExpression.create(
                    arg=time_comparison_column_expr, time_granularity=grain_to_date
                ),
            )
            comparison_expressions.append(start_of_range_comparison_expr)

        return SqlLogicalExpression.create(
            operator=SqlLogicalOperator.AND,
            args=tuple(comparison_expressions),
        )

    @staticmethod
    def make_join_conversion_join_description(
        node: JoinConversionEventsNode,
        base_data_set: AnnotatedSqlDataSet,
        conversion_data_set: AnnotatedSqlDataSet,
        column_equality_descriptions: Sequence[ColumnEqualityDescription],
    ) -> SqlJoinDescription:
        """Make a join description for joining base-to-conversion data sets.

        A successful conversion is when a conversion event occurs after a corresponding base event
        under the condition of being within the time range and other conditions (such as constant properties).
        This builds the join description to satisfy all those conditions.
        """
        window_condition = SqlPlanJoinBuilder._make_time_range_window_join_condition(
            base_data_set=base_data_set,
            time_comparison_dataset=conversion_data_set,
            window=node.window,
        )
        return SqlPlanJoinBuilder.make_column_equality_sql_join_description(
            right_source_node=conversion_data_set.data_set.checked_sql_select_node,
            left_source_alias=base_data_set.alias,
            right_source_alias=conversion_data_set.alias,
            column_equality_descriptions=column_equality_descriptions,
            join_type=SqlJoinType.INNER,
            additional_on_conditions=(window_condition,),
        )

    @staticmethod
    def make_cumulative_metric_time_range_join_description(
        node: JoinOverTimeRangeNode,
        metric_data_set: AnnotatedSqlDataSet,
        time_spine_data_set: AnnotatedSqlDataSet,
    ) -> SqlJoinDescription:
        """Make a join description to connect a cumulative metric input to a time spine dataset.

        Cumulative metrics must be joined against a time spine in a backward-looking fashion, with
        a range determined by a time window (delta against metric_time) and optional cumulative grain.
        """
        cumulative_join_condition = SqlPlanJoinBuilder._make_time_range_window_join_condition(
            base_data_set=metric_data_set,
            time_comparison_dataset=time_spine_data_set,
            window=node.window,
            grain_to_date=node.grain_to_date,
        )

        return SqlJoinDescription(
            right_source=metric_data_set.data_set.checked_sql_select_node,
            right_source_alias=metric_data_set.alias,
            on_condition=cumulative_join_condition,
            join_type=SqlJoinType.INNER,
        )

    @staticmethod
    def make_join_to_time_spine_join_description(
        node: JoinToTimeSpineNode,
        time_spine_alias: str,
        agg_time_dimension_column_name: str,
        parent_sql_select_node: SqlSelectStatementNode,
        parent_alias: str,
    ) -> SqlJoinDescription:
        """Build join expression used to join a metric to a time spine dataset."""
        left_expr: SqlExpressionNode = SqlColumnReferenceExpression.create(
            col_ref=SqlColumnReference(table_alias=time_spine_alias, column_name=agg_time_dimension_column_name)
        )
        if node.offset_window:
            left_expr = SqlSubtractTimeIntervalExpression.create(
                arg=left_expr, count=node.offset_window.count, granularity=node.offset_window.granularity
            )
        elif node.offset_to_grain:
            left_expr = SqlDateTruncExpression.create(time_granularity=node.offset_to_grain, arg=left_expr)

        return SqlJoinDescription(
            right_source=parent_sql_select_node,
            right_source_alias=parent_alias,
            on_condition=SqlComparisonExpression.create(
                left_expr=left_expr,
                comparison=SqlComparison.EQUALS,
                right_expr=SqlColumnReferenceExpression.create(
                    col_ref=SqlColumnReference(table_alias=parent_alias, column_name=agg_time_dimension_column_name)
                ),
            ),
            join_type=node.join_type,
        )



from __future__ import annotations

from dataclasses import dataclass

from dbt_semantic_interfaces.protocols import MetricInput
from dbt_semantic_interfaces.references import MetricReference

from metricflow_semantics.model.semantics.metric_lookup import MetricLookup


@dataclass(frozen=True)
class InputMetricDefinitionLocation:
    """Describes the location of input metric of a derived metric."""

    derived_metric_reference: MetricReference
    input_metric_list_index: int

    def get_metric_input(self, metric_lookup: MetricLookup) -> MetricInput:
        """Get the associated MetricInput object that this describes."""
        metric = metric_lookup.get_metric(self.derived_metric_reference)
        if metric.input_metrics is None or self.input_metric_list_index >= len(metric.input_metrics):
            raise ValueError(f"The metric input index is invalid for metric: {metric}")

        return metric.input_metrics[self.input_metric_list_index]



from __future__ import annotations

import itertools
import logging
from dataclasses import dataclass
from typing import Sequence, Tuple

from dbt_semantic_interfaces.references import SemanticModelReference
from typing_extensions import override

from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.model.semantics.linkable_element_set import LinkableElementSet
from metricflow_semantics.query.group_by_item.path_prefixable import PathPrefixable
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class GroupByItemCandidateSet(PathPrefixable, SemanticModelDerivation):
    """The set of candidate specs that could match a given spec pattern.

    This candidate set is refined as it is passed from the root node (representing a measure) to the leaf node
    (representing a query) in the resolution DAG.

    For logging / error reporting:

    * This includes the path from the leaf node to the measure node where the specs
    originated. There can be multiple paths as when candidate sets are merged, there can be multiple measures were a
    spec could have come from.

    * This includes the path from the leaf node to the node where this set was generated since during debugging or in
    error messages, you start analyzing from the leaf node.
    """

    linkable_element_set: LinkableElementSet
    measure_paths: Tuple[MetricFlowQueryResolutionPath, ...]
    path_from_leaf_node: MetricFlowQueryResolutionPath

    def __post_init__(self) -> None:  # noqa: D105
        # If there are no specs, there shouldn't be any measure paths.
        assert (len(self.specs) > 0 and len(self.measure_paths) > 0) or (
            len(self.specs) == 0 and len(self.measure_paths) == 0
        )

    @property
    def specs(self) -> Sequence[LinkableInstanceSpec]:  # noqa: D102
        return self.linkable_element_set.specs

    @staticmethod
    def intersection(
        path_from_leaf_node: MetricFlowQueryResolutionPath, candidate_sets: Sequence[GroupByItemCandidateSet]
    ) -> GroupByItemCandidateSet:
        """Create a new candidate set that is the intersection of the given candidate sets.

        The intersection is defined as the specs common to all candidate sets. path_from_leaf_node is used to indicate
        where the new candidate set was created.
        """
        if len(candidate_sets) == 0:
            return GroupByItemCandidateSet.empty_instance()
        elif len(candidate_sets) == 1:
            return GroupByItemCandidateSet(
                linkable_element_set=candidate_sets[0].linkable_element_set,
                measure_paths=candidate_sets[0].measure_paths,
                path_from_leaf_node=path_from_leaf_node,
            )
        linkable_element_set_candidates = tuple(candidate_set.linkable_element_set for candidate_set in candidate_sets)
        intersection_result = LinkableElementSet.intersection_by_path_key(linkable_element_set_candidates)
        if intersection_result.spec_count == 0:
            return GroupByItemCandidateSet.empty_instance()

        measure_paths = tuple(
            itertools.chain.from_iterable(candidate_set.measure_paths for candidate_set in candidate_sets)
        )

        return GroupByItemCandidateSet(
            linkable_element_set=intersection_result,
            measure_paths=measure_paths,
            path_from_leaf_node=path_from_leaf_node,
        )

    @property
    def is_empty(self) -> bool:  # noqa: D102
        return len(self.specs) == 0

    @property
    def num_candidates(self) -> int:  # noqa: D102
        return len(self.specs)

    @staticmethod
    def empty_instance() -> GroupByItemCandidateSet:  # noqa: D102
        return GroupByItemCandidateSet(
            linkable_element_set=LinkableElementSet(),
            measure_paths=(),
            path_from_leaf_node=MetricFlowQueryResolutionPath.empty_instance(),
        )

    def filter_candidates_by_pattern(
        self,
        spec_pattern: SpecPattern,
    ) -> GroupByItemCandidateSet:
        """Return a new candidate set that only contains specs that match the given pattern."""
        filtered_element_set = self.linkable_element_set.filter_by_spec_patterns((spec_pattern,))
        if filtered_element_set.spec_count == 0:
            return GroupByItemCandidateSet.empty_instance()
        return GroupByItemCandidateSet(
            linkable_element_set=filtered_element_set,
            measure_paths=self.measure_paths,
            path_from_leaf_node=self.path_from_leaf_node,
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> GroupByItemCandidateSet:
        return GroupByItemCandidateSet(
            linkable_element_set=self.linkable_element_set,
            measure_paths=tuple(path.with_path_prefix(path_prefix) for path in self.measure_paths),
            path_from_leaf_node=self.path_from_leaf_node.with_path_prefix(path_prefix),
        )

    @property
    @override
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        return self.linkable_element_set.derived_from_semantic_models



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence, Set

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor
from metricflow.dataflow.nodes.compute_metrics import ComputeMetricsNode


@dataclass(frozen=True, eq=False)
class WindowReaggregationNode(DataflowPlanNode):
    """A node that re-aggregates metrics using window functions.

    Currently used for calculating cumulative metrics at various granularities.

    Attributes:
        metric_spec: Specification of the metric to be re-aggregated.
        order_by_spec: Specification of the time dimension to order by.
        partition_by_specs: Specifications of the instances to partition by.
    """

    metric_spec: MetricSpec
    order_by_spec: TimeDimensionSpec
    partition_by_specs: Sequence[InstanceSpec]

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1
        if self.order_by_spec in self.partition_by_specs:
            raise ValueError(
                "Order by spec found in partition by specs for WindowAggregationNode. This indicates internal misconfiguration"
                f" because reaggregation should not be needed in this circumstance. Order by spec: {self.order_by_spec}; "
                f"Partition by specs:{self.partition_by_specs}"
            )

    @staticmethod
    def create(  # noqa: D102
        parent_node: ComputeMetricsNode,
        metric_spec: MetricSpec,
        order_by_spec: TimeDimensionSpec,
        partition_by_specs: Sequence[InstanceSpec],
    ) -> WindowReaggregationNode:
        return WindowReaggregationNode(
            parent_nodes=(parent_node,),
            metric_spec=metric_spec,
            order_by_spec=order_by_spec,
            partition_by_specs=partition_by_specs,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_WINDOW_REAGGREGATION_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_window_reaggregation_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Re-aggregate Metrics via Window Functions"""

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (
            DisplayedProperty("metric_spec", self.metric_spec),
            DisplayedProperty("order_by_spec", self.order_by_spec),
            DisplayedProperty("partition_by_specs", self.partition_by_specs),
        )

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.parent_nodes == self.parent_nodes
            and other_node.metric_spec == self.metric_spec
            and other_node.order_by_spec == self.order_by_spec
            and other_node.partition_by_specs == self.partition_by_specs
        )

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> WindowReaggregationNode:  # noqa: D102
        assert len(new_parent_nodes) == 1, "WindowReaggregationNode cannot accept multiple parents."
        new_parent_node = new_parent_nodes[0]
        assert isinstance(
            new_parent_node, ComputeMetricsNode
        ), "WindowReaggregationNode can only have ComputeMetricsNode as parent node."
        return WindowReaggregationNode.create(
            parent_node=new_parent_node,
            metric_spec=self.metric_spec,
            order_by_spec=self.order_by_spec,
            partition_by_specs=self.partition_by_specs,
        )

    @property
    def aggregated_to_elements(self) -> Set[LinkableInstanceSpec]:  # noqa: D102
        return set()



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Sequence

from dbt_semantic_interfaces.call_parameter_sets import (
    DimensionCallParameterSet,
    EntityCallParameterSet,
    MetricCallParameterSet,
    TimeDimensionCallParameterSet,
)
from dbt_semantic_interfaces.references import EntityReference
from typing_extensions import override

from metricflow_semantics.model.linkable_element_property import LinkableElementProperty
from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.patterns.entity_link_pattern import (
    EntityLinkPattern,
    ParameterSetField,
    SpecPatternParameterSet,
)
from metricflow_semantics.specs.spec_set import group_specs_by_type


@dataclass(frozen=True)
class DimensionPattern(EntityLinkPattern):
    """Similar to EntityPathPattern but only matches dimensions / time dimensions.

    Analogous pattern for Dimension() in the object builder naming scheme.
    """

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[LinkableInstanceSpec]:
        spec_set = group_specs_by_type(candidate_specs)
        filtered_specs: Sequence[LinkableInstanceSpec] = spec_set.dimension_specs + spec_set.time_dimension_specs
        return super().match(filtered_specs)

    @staticmethod
    def from_call_parameter_set(  # noqa: D102
        dimension_call_parameter_set: DimensionCallParameterSet,
    ) -> DimensionPattern:
        return DimensionPattern(
            parameter_set=SpecPatternParameterSet.from_parameters(
                fields_to_compare=(
                    ParameterSetField.ELEMENT_NAME,
                    ParameterSetField.ENTITY_LINKS,
                ),
                element_name=dimension_call_parameter_set.dimension_reference.element_name,
                entity_links=dimension_call_parameter_set.entity_path,
            )
        )

    @property
    @override
    def element_pre_filter(self) -> LinkableElementFilter:
        return super().element_pre_filter.merge(
            LinkableElementFilter(without_any_of=frozenset({LinkableElementProperty.METRIC}))
        )


@dataclass(frozen=True)
class TimeDimensionPattern(EntityLinkPattern):
    """Similar to EntityPathPattern but only matches time dimensions.

    Analogous pattern for TimeDimension() in the object builder naming scheme.
    """

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[LinkableInstanceSpec]:
        spec_set = group_specs_by_type(candidate_specs)
        return super().match(spec_set.time_dimension_specs)

    @staticmethod
    def from_call_parameter_set(
        time_dimension_call_parameter_set: TimeDimensionCallParameterSet,
    ) -> TimeDimensionPattern:
        """Create the pattern that represents 'TimeDimension(...)' in the object builder naming scheme.

        For this pattern, A None value for the time grain matches any grain. However, a None value for the date part
        means that the date part has to be None. This follows the interface defined by the object builder naming scheme.
        """
        fields_to_compare: List[ParameterSetField] = [
            ParameterSetField.ELEMENT_NAME,
            ParameterSetField.ENTITY_LINKS,
            ParameterSetField.DATE_PART,
        ]

        if time_dimension_call_parameter_set.time_granularity_name is not None:
            fields_to_compare.append(ParameterSetField.TIME_GRANULARITY)

        return TimeDimensionPattern(
            parameter_set=SpecPatternParameterSet.from_parameters(
                fields_to_compare=tuple(fields_to_compare),
                element_name=time_dimension_call_parameter_set.time_dimension_reference.element_name,
                entity_links=time_dimension_call_parameter_set.entity_path,
                time_granularity_name=time_dimension_call_parameter_set.time_granularity_name,
                date_part=time_dimension_call_parameter_set.date_part,
            )
        )

    @property
    @override
    def element_pre_filter(self) -> LinkableElementFilter:
        return super().element_pre_filter.merge(
            LinkableElementFilter(without_any_of=frozenset({LinkableElementProperty.METRIC}))
        )


@dataclass(frozen=True)
class EntityPattern(EntityLinkPattern):
    """Similar to EntityPathPattern but only matches entities.

    Analogous pattern for Entity() in the object builder naming scheme.
    """

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[LinkableInstanceSpec]:
        spec_set = group_specs_by_type(candidate_specs)
        return super().match(spec_set.entity_specs)

    @staticmethod
    def from_call_parameter_set(entity_call_parameter_set: EntityCallParameterSet) -> EntityPattern:  # noqa: D102
        return EntityPattern(
            parameter_set=SpecPatternParameterSet.from_parameters(
                fields_to_compare=(
                    ParameterSetField.ELEMENT_NAME,
                    ParameterSetField.ENTITY_LINKS,
                ),
                element_name=entity_call_parameter_set.entity_reference.element_name,
                entity_links=entity_call_parameter_set.entity_path,
            )
        )

    @property
    @override
    def element_pre_filter(self) -> LinkableElementFilter:
        return LinkableElementFilter(without_any_of=frozenset({LinkableElementProperty.METRIC}))


@dataclass(frozen=True)
class GroupByMetricPattern(EntityLinkPattern):
    """A pattern that matches metrics using the group by specifications."""

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[LinkableInstanceSpec]:
        spec_set = group_specs_by_type(candidate_specs)
        return super().match(spec_set.group_by_metric_specs)

    @staticmethod
    def from_call_parameter_set(  # noqa: D102
        metric_call_parameter_set: MetricCallParameterSet,
    ) -> GroupByMetricPattern:
        # This looks hacky because the typing for the interface does not match the implementation, but that's temporary!
        # This will get a lot less hacky once we enable multiple entities and dimensions in the group by.
        if len(metric_call_parameter_set.group_by) != 1:
            raise RuntimeError(
                "Currently only one group by item is allowed for Metric filters. "
                "This should have been caught by validations."
            )
        group_by = metric_call_parameter_set.group_by[0]
        # custom_granularity_names is empty because we are not parsing any dimensions here with grain
        structured_name = StructuredLinkableSpecName.from_name(
            qualified_name=group_by.element_name, custom_granularity_names=()
        )
        metric_subquery_entity_links = tuple(
            EntityReference(entity_name)
            for entity_name in (structured_name.entity_link_names + (structured_name.element_name,))
        )
        # Temp: we don't have a parameter to specify the join path from the outer query to the metric subquery,
        # so just use the last entity. Will need to add another param for that later.
        entity_links = metric_subquery_entity_links[-1:]
        return GroupByMetricPattern(
            parameter_set=SpecPatternParameterSet.from_parameters(
                fields_to_compare=(
                    ParameterSetField.ELEMENT_NAME,
                    ParameterSetField.ENTITY_LINKS,
                    ParameterSetField.METRIC_SUBQUERY_ENTITY_LINKS,
                ),
                element_name=metric_call_parameter_set.metric_reference.element_name,
                entity_links=entity_links,
                metric_subquery_entity_links=metric_subquery_entity_links,
            )
        )



from __future__ import annotations

from dbt_semantic_interfaces.implementations.semantic_manifest import (
    PydanticSemanticManifest,
)
from dbt_semantic_interfaces.transformations.boolean_measure import (
    BooleanMeasureAggregationRule,
)
from dbt_semantic_interfaces.transformations.convert_count import ConvertCountToSumRule
from dbt_semantic_interfaces.transformations.convert_median import (
    ConvertMedianToPercentileRule,
)
from dbt_semantic_interfaces.transformations.cumulative_type_params import SetCumulativeTypeParamsRule
from dbt_semantic_interfaces.transformations.names import LowerCaseNamesRule
from dbt_semantic_interfaces.transformations.proxy_measure import CreateProxyMeasureRule
from dbt_semantic_interfaces.transformations.semantic_manifest_transformer import (
    PydanticSemanticManifestTransformer,
)

from metricflow_semantics.model.transformations.dedupe_metric_input_measures import DedupeMetricInputMeasuresRule


def parse_manifest_from_dbt_generated_manifest(manifest_json_string: str) -> PydanticSemanticManifest:
    """Parse a PydanticSemanticManifest given the generated semantic_manifest json from dbt."""
    raw_model = PydanticSemanticManifest.parse_raw(manifest_json_string)
    # The serialized object in the dbt project does not have all transformations applied to it at
    # this time, which causes failures with input measure resolution.
    # TODO: remove this transform call once the upstream changes are integrated into our dependency tree
    # TODO: align rules between DSI, here, and MFS (if possible!)
    rules = (
        # Primary
        (LowerCaseNamesRule(),),
        # Secondary
        (
            CreateProxyMeasureRule(),
            BooleanMeasureAggregationRule(),
            ConvertCountToSumRule(),
            ConvertMedianToPercentileRule(),
            DedupeMetricInputMeasuresRule(),  # Remove once fix is in core
            SetCumulativeTypeParamsRule(),
        ),
    )
    model = PydanticSemanticManifestTransformer.transform(raw_model, rules)
    return model



from __future__ import annotations

from dataclasses import dataclass
from typing import FrozenSet, Optional

from typing_extensions import Self, override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.model.linkable_element_property import LinkableElementProperty


@dataclass(frozen=True)
class LinkableElementFilter(Mergeable):
    """Describes a way to filter the `LinkableElements` in a `LinkableElementSet`."""

    # A `None` value for element names means no filtering on element names.
    element_names: Optional[FrozenSet[str]] = None
    with_any_of: FrozenSet[LinkableElementProperty] = LinkableElementProperty.all_properties()
    without_any_of: FrozenSet[LinkableElementProperty] = frozenset()
    without_all_of: FrozenSet[LinkableElementProperty] = frozenset()

    @override
    def merge(self: Self, other: LinkableElementFilter) -> LinkableElementFilter:
        if self.element_names is None and other.element_names is None:
            element_names = None
        else:
            element_names = (self.element_names or frozenset()).union(other.element_names or frozenset())
        return LinkableElementFilter(
            element_names=element_names,
            with_any_of=self.with_any_of.union(other.with_any_of),
            without_any_of=self.without_any_of.union(other.without_any_of),
            without_all_of=self.without_all_of.union(other.without_all_of),
        )

    @classmethod
    @override
    def empty_instance(cls) -> LinkableElementFilter:
        return LinkableElementFilter()

    def without_element_names(self) -> LinkableElementFilter:
        """Return this filter without the `element_names` filter set."""
        return LinkableElementFilter(
            with_any_of=self.with_any_of,
            without_any_of=self.without_any_of,
            without_all_of=self.without_all_of,
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Sequence, Tuple

from dbt_semantic_interfaces.references import MetricReference
from more_itertools import is_sorted


class WhereFilterLocationType(Enum):
    """Describes the location where the filter is defined."""

    QUERY = "query"
    METRIC = "metric"
    INPUT_METRIC = "input_metric"


@dataclass(frozen=True)
class WhereFilterLocation:
    """Describe the location of a where filter.

    When describing a query, the metric references are the metrics in the query.
    When describing a metric, the metric reference just the metric where the filter is defined.
    When describing an input metric, the metric references are the outer metric and the input metric.

    Since this is used to resolve valid group-by items for a where filter, this is sufficient as the valid group-by
    items depend only on the metrics queried.
    """

    location_type: WhereFilterLocationType
    # These should be sorted for consistency in comparisons.
    metric_references: Tuple[MetricReference, ...]

    def __post_init__(self) -> None:  # noqa: D105
        assert is_sorted(self.metric_references)

    @staticmethod
    def for_query(metric_references: Sequence[MetricReference]) -> WhereFilterLocation:  # noqa: D102
        return WhereFilterLocation(
            metric_references=tuple(sorted(metric_references)), location_type=WhereFilterLocationType.QUERY
        )

    @staticmethod
    def for_metric(metric_reference: MetricReference) -> WhereFilterLocation:  # noqa: D102
        return WhereFilterLocation(metric_references=(metric_reference,), location_type=WhereFilterLocationType.METRIC)

    @staticmethod
    def for_input_metric(input_metric_reference: MetricReference) -> WhereFilterLocation:  # noqa: D102
        return WhereFilterLocation(
            metric_references=(input_metric_reference,), location_type=WhereFilterLocationType.INPUT_METRIC
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.specs.order_by_spec import OrderBySpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class OrderByLimitNode(DataflowPlanNode):
    """A node that re-orders the input data with a limit.

    Attributes:
        order_by_specs: Describes how to order the incoming data.
        limit: Number of rows to limit.
    """

    order_by_specs: Sequence[OrderBySpec]
    limit: Optional[int]

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        order_by_specs: Sequence[OrderBySpec],
        parent_node: DataflowPlanNode,
        limit: Optional[int] = None,
    ) -> OrderByLimitNode:
        return OrderByLimitNode(
            parent_nodes=(parent_node,),
            order_by_specs=tuple(order_by_specs),
            limit=limit,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_ORDER_BY_LIMIT_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_order_by_limit_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"Order By {[order_by_spec.instance_spec.qualified_name for order_by_spec in self.order_by_specs]}" + (
            f" Limit {self.limit}" if self.limit else ""
        )

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return (
            tuple(super().displayed_properties)
            + tuple(DisplayedProperty("order_by_spec", order_by_spec) for order_by_spec in self.order_by_specs)
            + (DisplayedProperty("limit", str(self.limit)),)
        )

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.order_by_specs == self.order_by_specs
            and other_node.limit == self.limit
        )

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> OrderByLimitNode:  # noqa: D102
        assert len(new_parent_nodes) == 1

        return OrderByLimitNode.create(
            parent_node=new_parent_nodes[0],
            order_by_specs=self.order_by_specs,
            limit=self.limit,
        )



from __future__ import annotations

import logging
from typing import Sequence, Tuple

from metricflow_semantics.naming.naming_scheme import QueryItemNamingScheme
from metricflow_semantics.query.similarity import top_fuzzy_matches
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.patterns.minimum_time_grain import MinimumTimeGrainPattern
from metricflow_semantics.specs.patterns.no_group_by_metric import NoGroupByMetricPattern
from metricflow_semantics.specs.patterns.none_date_part import NoneDatePartPattern
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern

logger = logging.getLogger(__name__)


class QueryItemSuggestionGenerator:
    """Returns specs that partially match a spec pattern created from user input. Used for suggestions in errors.

    Since suggestions are needed for group-by-items specified in the query and in where filters, an optional candidate
    filter can be specified to limit suggestions to the ones valid for the entire query. For use with where filters,
    a candidate filter is not needed as any available spec at a resolution node can be used.
    """

    # Adding these filters so that we don't get multiple suggestions that are similar, but with different
    # grains. Some additional thought is needed to tweak this as the base grain may not be the best suggestion.
    FILTER_ITEM_CANDIDATE_FILTERS: Tuple[SpecPattern, ...] = (MinimumTimeGrainPattern(), NoneDatePartPattern())
    GROUP_BY_ITEM_CANDIDATE_FILTERS: Tuple[SpecPattern, ...] = (
        MinimumTimeGrainPattern(),
        NoneDatePartPattern(),
        NoGroupByMetricPattern(),
    )

    def __init__(  # noqa: D107
        self,
        input_naming_scheme: QueryItemNamingScheme,
        input_str: str,
        candidate_filters: Sequence[SpecPattern],
    ) -> None:
        self._input_naming_scheme = input_naming_scheme
        self._input_str = input_str
        self._candidate_filters = candidate_filters

    @property
    def candidate_filters(self) -> Sequence[SpecPattern]:
        """Return the filters that should be applied to the candidate specs when generating suggestions."""
        return self._candidate_filters

    def input_suggestions(
        self,
        candidate_specs: Sequence[InstanceSpec],
        max_suggestions: int = 6,
    ) -> Sequence[str]:
        """Return the best specs that match the given pattern from candidate_specs and match the candidate_filer."""
        # Use edit distance to figure out the closest matches, so convert the specs to strings.

        for candidate_filter in self._candidate_filters:
            candidate_specs = candidate_filter.match(candidate_specs)

        candidate_strs = set()
        for candidate_spec in candidate_specs:
            candidate_str = self._input_naming_scheme.input_str(candidate_spec)

            if candidate_str is not None:
                candidate_strs.add(candidate_str)

        fuzzy_matches = top_fuzzy_matches(
            item=self._input_str,
            candidate_items=tuple(candidate_strs),
            max_matches=max_suggestions,
        )

        return tuple(scored_item.item_str for scored_item in fuzzy_matches)



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence, Tuple

from typing_extensions import override

from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern


@dataclass(frozen=True)
class MatchListSpecPattern(SpecPattern):
    """A spec pattern that matches based on a configured list of specs.

    This is useful for filtering possible group-by-items to ones valid for a query.
    """

    listed_specs: Tuple[InstanceSpec, ...]

    @staticmethod
    def create(listed_specs: Sequence[InstanceSpec]) -> MatchListSpecPattern:  # noqa: D102
        return MatchListSpecPattern(tuple(listed_specs))

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[InstanceSpec]:
        return tuple(spec for spec in candidate_specs if spec in self.listed_specs)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from dbt_semantic_interfaces.protocols import ProtocolHint
from dbt_semantic_interfaces.references import EntityReference
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from typing_extensions import override

from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.naming.metric_scheme import MetricNamingScheme
from metricflow_semantics.naming.object_builder_scheme import ObjectBuilderNamingScheme
from metricflow_semantics.protocols.query_parameter import (
    DimensionOrEntityQueryParameter,
    InputOrderByParameter,
    MetricQueryParameter,
    OrderByQueryParameter,
    TimeDimensionQueryParameter,
)
from metricflow_semantics.protocols.query_parameter import SavedQueryParameter as SavedQueryParameterProtocol
from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import (
    ResolverInputForGroupByItem,
    ResolverInputForMetric,
    ResolverInputForOrderByItem,
)
from metricflow_semantics.specs.patterns.entity_link_pattern import (
    EntityLinkPattern,
    ParameterSetField,
    SpecPatternParameterSet,
)


@dataclass(frozen=True)
class TimeDimensionParameter(ProtocolHint[TimeDimensionQueryParameter]):
    """Time dimension requested in a query."""

    def _implements_protocol(self) -> TimeDimensionQueryParameter:
        return self

    name: str
    grain: Optional[str] = None
    date_part: Optional[DatePart] = None

    def query_resolver_input(  # noqa: D102
        self,
        semantic_manifest_lookup: SemanticManifestLookup,
    ) -> ResolverInputForGroupByItem:
        fields_to_compare = [
            ParameterSetField.ELEMENT_NAME,
            ParameterSetField.ENTITY_LINKS,
            ParameterSetField.DATE_PART,
        ]
        if self.grain is not None:
            fields_to_compare.append(ParameterSetField.TIME_GRANULARITY)

        name_structure = StructuredLinkableSpecName.from_name(
            qualified_name=self.name.lower(),
            custom_granularity_names=semantic_manifest_lookup.semantic_model_lookup.custom_granularity_names,
        )

        return ResolverInputForGroupByItem(
            input_obj=self,
            input_obj_naming_scheme=ObjectBuilderNamingScheme(),
            spec_pattern=EntityLinkPattern(
                SpecPatternParameterSet.from_parameters(
                    fields_to_compare=tuple(fields_to_compare),
                    element_name=name_structure.element_name,
                    entity_links=tuple(EntityReference(link_name) for link_name in name_structure.entity_link_names),
                    time_granularity_name=self.grain,
                    date_part=self.date_part,
                )
            ),
        )


@dataclass(frozen=True)
class DimensionOrEntityParameter(ProtocolHint[DimensionOrEntityQueryParameter]):
    """Group by parameter requested in a query.

    Might represent an entity or a dimension.
    """

    name: str

    @override
    def _implements_protocol(self) -> DimensionOrEntityQueryParameter:
        return self

    def query_resolver_input(self, semantic_manifest_lookup: SemanticManifestLookup) -> ResolverInputForGroupByItem:
        """Produces resolver input from a query parameter representing a dimension or entity.

        Note these parameters do not currently have a direct need for the semantic_manifest_lookup, but since these
        can be lumped in with other items that do require it we keep this method signature consistent across
        the class sets.

        TODO: Refine these query input classes so that this kind of thing is either enforced in self-documenting
        ways or removed from the codebase
        """
        name_structure = StructuredLinkableSpecName.from_name(
            qualified_name=self.name.lower(),
            custom_granularity_names=semantic_manifest_lookup.semantic_model_lookup.custom_granularity_names,
        )

        return ResolverInputForGroupByItem(
            input_obj=self,
            input_obj_naming_scheme=ObjectBuilderNamingScheme(),
            spec_pattern=EntityLinkPattern(
                SpecPatternParameterSet.from_parameters(
                    fields_to_compare=(
                        ParameterSetField.ELEMENT_NAME,
                        ParameterSetField.ENTITY_LINKS,
                        ParameterSetField.DATE_PART,
                    ),
                    element_name=name_structure.element_name,
                    entity_links=tuple(EntityReference(link_name) for link_name in name_structure.entity_link_names),
                    time_granularity_name=None,
                    date_part=None,
                )
            ),
        )


@dataclass(frozen=True)
class MetricParameter(ProtocolHint[MetricQueryParameter]):
    """Metric requested in a query."""

    name: str
    alias: Optional[str] = None

    @override
    def _implements_protocol(self) -> MetricQueryParameter:
        return self

    def query_resolver_input(  # noqa: D102
        self, semantic_manifest_lookup: SemanticManifestLookup
    ) -> ResolverInputForMetric:
        naming_scheme = MetricNamingScheme()
        return ResolverInputForMetric(
            input_obj=self,
            naming_scheme=naming_scheme,
            spec_pattern=naming_scheme.spec_pattern(self.name, semantic_manifest_lookup=semantic_manifest_lookup),
            alias=self.alias,
        )


@dataclass(frozen=True)
class OrderByParameter(ProtocolHint[OrderByQueryParameter]):
    """Order by requested in a query."""

    order_by: InputOrderByParameter
    descending: bool = False

    @override
    def _implements_protocol(self) -> OrderByQueryParameter:
        return self

    def query_resolver_input(  # noqa: D102
        self, semantic_manifest_lookup: SemanticManifestLookup
    ) -> ResolverInputForOrderByItem:
        return ResolverInputForOrderByItem(
            input_obj=self,
            possible_inputs=(self.order_by.query_resolver_input(semantic_manifest_lookup=semantic_manifest_lookup),),
            descending=self.descending,
        )


@dataclass(frozen=True)
class SavedQueryParameter(ProtocolHint[SavedQueryParameterProtocol]):
    """Dataclass implementation of SavedQueryParameterProtocol."""

    name: str

    @override
    def _implements_protocol(self) -> SavedQueryParameterProtocol:
        return self



from __future__ import annotations

from abc import ABC, abstractmethod

from typing_extensions import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath


class PathPrefixable(ABC):
    """Describes an object that contains a path that can be updated with a prefix path.

    During a recursive traversal of a DAG, we keep track of the path from the start node to the current node. From the
    current node, a method generates a relative path from the current node to a target node. To get the path from the
    start node to the target node, those to paths can be joined with the path from the start node to the current node
    as a prefix of the path from the current node to the target node.
    """

    @abstractmethod
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> Self:
        """Return a copy of Self, but with the associated path to include the path_prefix_node at the beginning."""
        raise NotImplementedError



from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, Tuple

from typing_extensions import override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.dag.mf_dag import DagNode
from metricflow_semantics.visitor import Visitable, VisitorOutputT

if TYPE_CHECKING:
    from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.measure_source_node import (
        MeasureGroupByItemSourceNode,
    )
    from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.metric_resolution_node import (
        MetricGroupByItemResolutionNode,
    )
    from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.no_metrics_query_source_node import (
        NoMetricsGroupByItemSourceNode,
    )
    from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.query_resolution_node import (
        QueryGroupByItemResolutionNode,
    )


@dataclass(frozen=True)
class GroupByItemResolutionNode(DagNode["GroupByItemResolutionNode"], Visitable, ABC):
    """Base node type for nodes in a GroupByItemResolutionDag.

    See GroupByItemResolutionDag for more details.
    """

    parent_nodes: Tuple[GroupByItemResolutionNode, ...]

    @abstractmethod
    def accept(self, visitor: GroupByItemResolutionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        """Called when a visitor needs to visit this node."""
        raise NotImplementedError

    @property
    @abstractmethod
    def ui_description(self) -> str:
        """A string that can be used to describe this node as a path element in the UI."""
        raise NotImplementedError

    @abstractmethod
    def _self_set(self) -> GroupByItemResolutionNodeSet:
        """Return a `GroupByItemResolutionNodeInclusiveAncestorSet` only containing self.

        Use to simplify implementation of `inclusive_ancestors`
        """
        raise NotImplementedError

    def inclusive_ancestors(self) -> GroupByItemResolutionNodeSet:
        """Return a set containing itself and all its ancestors."""
        return GroupByItemResolutionNodeSet.merge_iterable(
            itertools.chain(
                [self._self_set()], (parent_node.inclusive_ancestors() for parent_node in self.parent_nodes)
            )
        )


class GroupByItemResolutionNodeVisitor(Generic[VisitorOutputT], ABC):
    """Visitor for traversing GroupByItemResolutionNodes."""

    @abstractmethod
    def visit_measure_node(self, node: MeasureGroupByItemSourceNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_no_metrics_query_node(self, node: NoMetricsGroupByItemSourceNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_metric_node(self, node: MetricGroupByItemResolutionNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_query_node(self, node: QueryGroupByItemResolutionNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError


@dataclass(frozen=True)
class GroupByItemResolutionNodeSet(Mergeable):
    """Set containing nodes in a group-by-item resolution DAG."""

    measure_nodes: Tuple[MeasureGroupByItemSourceNode, ...] = ()
    no_metrics_query_nodes: Tuple[NoMetricsGroupByItemSourceNode, ...] = ()
    metric_nodes: Tuple[MetricGroupByItemResolutionNode, ...] = ()
    query_nodes: Tuple[QueryGroupByItemResolutionNode, ...] = ()

    @override
    def merge(self, other: GroupByItemResolutionNodeSet) -> GroupByItemResolutionNodeSet:
        return GroupByItemResolutionNodeSet(
            measure_nodes=self.measure_nodes + other.measure_nodes,
            no_metrics_query_nodes=self.no_metrics_query_nodes + other.no_metrics_query_nodes,
            metric_nodes=self.metric_nodes + other.metric_nodes,
            query_nodes=self.query_nodes + other.query_nodes,
        )

    @classmethod
    @override
    def empty_instance(cls) -> GroupByItemResolutionNodeSet:
        return GroupByItemResolutionNodeSet()



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Sequence, Tuple, Union

from dbt_semantic_interfaces.call_parameter_sets import (
    DimensionCallParameterSet,
    EntityCallParameterSet,
    MetricCallParameterSet,
    TimeDimensionCallParameterSet,
)
from dbt_semantic_interfaces.protocols import WhereFilterIntersection
from typing_extensions import override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.model.semantics.linkable_element import LinkableElement
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_location import WhereFilterLocation
from metricflow_semantics.query.group_by_item.path_prefixable import PathPrefixable
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import MetricFlowQueryResolutionIssueSet
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern

if TYPE_CHECKING:
    from metricflow_semantics.model.semantics.linkable_element_set import LinkableElementSet
    from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class FilterSpecResolutionLookUp(Mergeable):
    """Allows you to look up the spec for a group-by-item in a where filter.

    If there are issues parsing the filter, or if the filter had specified an invalid group-by-item, the appropriate
    issue is contained in the issue_set.
    """

    spec_resolutions: Tuple[FilterSpecResolution, ...]
    non_parsable_resolutions: Tuple[NonParsableFilterResolution, ...]

    @property
    def has_errors(self) -> bool:  # noqa: D102
        return any(
            non_parsable_resolution.issue_set.has_errors for non_parsable_resolution in self.non_parsable_resolutions
        ) or any(spec_resolution.issue_set.has_errors for spec_resolution in self.spec_resolutions)

    @property
    def has_issues(self) -> bool:  # noqa: D102
        return any(
            non_parsable_resolution.issue_set.has_issues for non_parsable_resolution in self.non_parsable_resolutions
        ) or any(spec_resolution.issue_set.has_issues for spec_resolution in self.spec_resolutions)

    def get_spec_resolutions(self, resolved_spec_lookup_key: ResolvedSpecLookUpKey) -> Sequence[FilterSpecResolution]:
        """Return the specs resolutions associated with the given key."""
        return tuple(
            resolution for resolution in self.spec_resolutions if resolution.lookup_key == resolved_spec_lookup_key
        )

    def spec_resolution_exists(self, resolved_spec_lookup_key: ResolvedSpecLookUpKey) -> bool:
        """Returns true if a resolution exists for the given key."""
        return len(self.get_spec_resolutions(resolved_spec_lookup_key)) > 0

    def _checked_resolution(self, resolved_spec_lookup_key: ResolvedSpecLookUpKey) -> FilterSpecResolution:
        """Helper to get just the resolution so we can access different properties on it."""
        resolutions = self.get_spec_resolutions(resolved_spec_lookup_key)
        if len(resolutions) == 0:
            raise RuntimeError(
                f"Unable to find a resolved spec.\n\n"
                f"Expected 1 resolution for:\n\n"
                f"{indent(mf_pformat(resolved_spec_lookup_key))}\n\n"
                f"but did not find any. All resolutions are:\n\n"
                f"{indent(mf_pformat(self.spec_resolutions))}"
            )

        # There may be multiple resolutions that match a given key because it's possible the same metric / filter is
        # used multiple times in a query (e.g. as a part of a derived metric). However, for a given metric and
        # a CallParameterSet, it should resolve to the same thing. Multiple resolutions are kept in this object for
        # error-reporting purposes.
        resolution = resolutions[0]
        if resolution.resolved_spec is None:
            raise RuntimeError(
                f"Expected a resolution with a resolved spec, but got:\n"
                f"{mf_pformat(resolution)}.\n"
                f"All resolutions are:\n"
                f"{mf_pformat(self.spec_resolutions)}"
            )

        return resolution

    def checked_resolved_spec(self, resolved_spec_lookup_key: ResolvedSpecLookUpKey) -> LinkableInstanceSpec:
        """Returns the resolved spec for the given key.

        If a resolution does not exist, or there is no spec associated with the resolution, this raises a RuntimeError.
        """
        resolution = self._checked_resolution(resolved_spec_lookup_key=resolved_spec_lookup_key)
        resolved_spec = resolution.resolved_spec
        assert (
            resolved_spec is not None
        ), f"Typechecker hint. Expected a resolution with a resolved spec, but got:\n{mf_pformat(resolution)}"
        return resolved_spec

    def checked_resolved_linkable_elements(
        self, resolved_spec_lookup_key: ResolvedSpecLookUpKey
    ) -> Sequence[LinkableElement]:
        """Returns the sequence of LinkableElements for the given spec lookup key.

        These are the LinkableElements bound to the singular spec/path_key for a given resolved filter item. They are
        useful for propagating metadata about the origin semantic model across the boundary between the filter resolver
        and the DataflowPlanBuilder.
        """
        return self._checked_resolution(resolved_spec_lookup_key=resolved_spec_lookup_key).resolved_linkable_elements

    @override
    def merge(self, other: FilterSpecResolutionLookUp) -> FilterSpecResolutionLookUp:
        return FilterSpecResolutionLookUp(
            spec_resolutions=self.spec_resolutions + other.spec_resolutions,
            non_parsable_resolutions=self.non_parsable_resolutions + other.non_parsable_resolutions,
        )

    @override
    @classmethod
    def empty_instance(cls) -> FilterSpecResolutionLookUp:
        return FilterSpecResolutionLookUp(
            spec_resolutions=(),
            non_parsable_resolutions=(),
        )


@dataclass(frozen=True)
class NonParsableFilterResolution(PathPrefixable):
    """A where filter intersection that couldn't be parsed e.g. Jinja error."""

    filter_location_path: MetricFlowQueryResolutionPath
    where_filter_intersection: WhereFilterIntersection
    issue_set: MetricFlowQueryResolutionIssueSet

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> NonParsableFilterResolution:
        return NonParsableFilterResolution(
            filter_location_path=self.filter_location_path.with_path_prefix(path_prefix),
            where_filter_intersection=self.where_filter_intersection,
            issue_set=self.issue_set.with_path_prefix(path_prefix),
        )


@dataclass(frozen=True)
class ResolvedSpecLookUpKey:
    """A key that associates a call parameter set and the filter where it is located."""

    filter_location: WhereFilterLocation
    call_parameter_set: CallParameterSet

    @staticmethod
    def from_parameters(  # noqa: D102
        filter_location: WhereFilterLocation, call_parameter_set: CallParameterSet
    ) -> ResolvedSpecLookUpKey:
        return ResolvedSpecLookUpKey(
            filter_location=filter_location,
            call_parameter_set=call_parameter_set,
        )


@dataclass(frozen=True)
class FilterSpecResolution:
    """Associates a lookup key with the resolved spec and its associated LinkableElements.

    We store the LinkableElementSet as a convenience container for managing the relationship between the
    singular resolved LinkableInstanceSpec and the collection of LinkableElements. This allows us to
    do things like fetch the full set of semantic models where the inputs for the given spec are defined.
    In most cases, there would be exactly one such semantic model, but in cases where, for example, an entity
    is used as a join key there might be multiple semantic model inputs for the given element, and we would
    need to be able to track that information.
    """

    lookup_key: ResolvedSpecLookUpKey
    where_filter_intersection: WhereFilterIntersection
    resolved_linkable_element_set: LinkableElementSet
    spec_pattern: SpecPattern
    issue_set: MetricFlowQueryResolutionIssueSet
    # Used for error messages.
    filter_location_path: MetricFlowQueryResolutionPath
    object_builder_str: str

    def __post_init__(self) -> None:
        """Validation to ensure there is exactly one resolved spec for a FilterSpecResolution.

        Due to the way the FilterSpecResolution is structured, the final output should contain a single spec.
        """
        num_specs = len(self.resolved_linkable_element_set.specs)
        assert num_specs <= 1, (
            f"Found {num_specs} in {self.resolved_linkable_element_set}, but a valid FilterSpecResolution should "
            "contain either 0 or 1 resolved specs."
        )

    @property
    def resolved_spec(self) -> Optional[LinkableInstanceSpec]:
        """Returns the lone resolved spec, if one was found.

        The final ValueError should not be reachable due to the post-init validation, but is in place in case someone
        updates or removes the latter without accounting for the possibility of runtime divergence.
        """
        specs = self.resolved_linkable_element_set.specs
        if len(specs) == 0:
            return None
        elif len(specs) == 1:
            return specs[0]
        else:
            raise ValueError(
                f"Found {len(specs)} in {self.resolved_linkable_element_set}, this should not be possible!"
            )

    @property
    def resolved_linkable_elements(self) -> Sequence[LinkableElement]:
        """Returns the resolved linkable elements, if any, for this resolution result."""
        resolved_spec = self.resolved_spec
        if resolved_spec is None:
            return tuple()

        return self.resolved_linkable_element_set.linkable_elements_for_path_key(resolved_spec.element_path_key)


CallParameterSet = Union[
    DimensionCallParameterSet, TimeDimensionCallParameterSet, EntityCallParameterSet, MetricCallParameterSet
]


@dataclass(frozen=True)
class PatternAssociationForWhereFilterGroupByItem:
    """Describes the pattern associated with a group-by-item in a where filter.

    e.g. "{{ TimeDimension('metric_time', 'day') }} = '2020-01-01'" ->
        GroupByItemInWhereFilter(
            call_parameter_set=TimeDimensionCallParameterSet('metric_time', DAY),
            input_str="TimeDimension('metric_time', 'day')",
            ...
        )
    """

    call_parameter_set: CallParameterSet
    object_builder_str: str
    spec_pattern: SpecPattern
